"""Tool shell prelude."""

# Due to typer usage:
# ruff: noqa: PLC0414, F401

from slurmbench.bash.items import Variable as BashVar
from slurmbench.samples.bash import get_sample_attribute as get_sample_attribute
from slurmbench.tool.bash import Argument as Argument
from slurmbench.tool.bash import OnlyOptions as OnlyOptions



"""Experiment bash steps."""

from enum import StrEnum


class PbfbenchDo(StrEnum):
    """Experiment bash steps."""

    SBATCH_COMMENTS = "SBATCH_COMMENTS"
    ARRAY_JOB_ID_FILE = "ARRAY_JOB_ID_FILE"
    STEP = "STEP"


class Steps(StrEnum):
    """Experiment bash steps."""

    INIT_ENV = "INIT_ENV"
    COMMAND = "COMMAND"
    CLOSE_ENV = "CLOSE_ENV"



"""Abstract tools results items module."""

from abc import ABC, abstractmethod

import slurmbench.experiment.file_system as exp_fs
import slurmbench.samples.items as smp
import slurmbench.samples.status as smp_status


class Result(ABC):
    """Result base."""

    def __init__(self, exp_fs_manager: exp_fs.ManagerBase) -> None:
        """Initialize."""
        self._exp_fs_manager = exp_fs_manager

    def exp_fs_manager(self) -> exp_fs.ManagerBase:
        """Get file system manager."""
        return self._exp_fs_manager

    # REFACTOR not sure it is relevant to use sample status, because of formatted
    @abstractmethod
    def check(self, sample: smp.RowNumbered) -> smp_status.Status:
        """Check input(s)."""
        raise NotImplementedError


class Original(Result):
    """Original result."""

    def check(self, sample: smp.RowNumbered) -> smp_status.Status:
        """Check input(s)."""
        return smp_status.get_status(
            self._exp_fs_manager.sample_fs_manager(sample),
        )


class Formatted(Result):
    """Formatted result."""



"""Tools environment logic.

Some tools require a specific environment,
such as a specific virtualenv, setting binary paths etc.

This module provides such logic.
"""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Self

if TYPE_CHECKING:
    from collections.abc import Iterator
    from pathlib import Path

_LOGGER = logging.getLogger(__name__)


class MissingMagicCommentError:
    """Missing magic comment error."""

    def __init__(self, script_path: Path, magic_comment: str) -> None:
        self._script_path = script_path
        self._magic_comment = magic_comment

    def script_path(self) -> Path:
        """Get the script path."""
        return self._script_path

    def magic_comment(self) -> str:
        """Get the magic comment."""
        return self._magic_comment

    def __str__(self) -> str:
        """Get string representation."""
        return (
            f"Missing magic comment `{self._magic_comment}`"
            f" in script `{self._script_path}`"
        )


class ScriptWrapper:
    """Wrapper to run a bash script in a specific environment."""

    BEGIN_ENV_MAGIC_COMMENT = "# PBFBENCH BEGIN_ENV"
    MID_ENV_MAGIC_COMMENT = "# PBFBENCH MID_ENV"
    END_ENV_MAGIC_COMMENT = "# PBFBENCH END_ENV"

    @classmethod
    def new(cls, script_path: Path) -> Self | MissingMagicCommentError:
        """Create a new environment wrapper."""
        match indices_or_err := cls._index_script(script_path):
            case MissingMagicCommentError():
                return indices_or_err
            case _:
                return cls(script_path, *indices_or_err)

    @classmethod
    def _index_script(
        cls,
        script_path: Path,
    ) -> tuple[int, int, int] | MissingMagicCommentError:
        """Index the script lines."""

        def _index_line_iter(script_path: Path) -> Iterator[tuple[int, str] | None]:
            """Iterate over the script lines."""
            with script_path.open("r") as f_in_script:
                line = next(f_in_script, None)
                k = 0
                while line is not None:
                    yield (k, line)
                    line = next(f_in_script, None)
                    k += 1
                yield None

        def _find_magic_comment(
            index_line_iter: Iterator[tuple[int, str] | None],
            magic_comment: str,
        ) -> int | MissingMagicCommentError:
            """Find magic comment or return error."""
            index_line = next(index_line_iter)
            while index_line is not None and not index_line[1].startswith(
                magic_comment,
            ):
                index_line = next(index_line_iter)
            if index_line is None:
                return MissingMagicCommentError(script_path, magic_comment)
            return index_line[0]

        index_line_iter = _index_line_iter(script_path)
        indices = []
        for magic_comment in (
            cls.BEGIN_ENV_MAGIC_COMMENT,
            cls.MID_ENV_MAGIC_COMMENT,
            cls.END_ENV_MAGIC_COMMENT,
        ):
            match index := _find_magic_comment(index_line_iter, magic_comment):
                case MissingMagicCommentError():
                    return index
                case int():
                    indices.append(index)
        return (indices[0], indices[1], indices[2])

    def __init__(
        self,
        script_path: Path,
        begin_line_index: int,
        mid_line_index: int,
        end_line_index: int,
    ) -> None:
        self.__script_path = script_path
        self.__begin_line_index = begin_line_index
        self.__mid_line_index = mid_line_index
        self.__end_line_index = end_line_index

    def init_env_lines(self) -> Iterator[str]:
        """Iterate over the script lines that init the environment."""
        with self.__script_path.open("r") as f_in_script:
            iter_lines = iter(f_in_script)
            line = next(iter_lines)
            k = 0
            # Skip before the begin magic comment
            while k < self.__begin_line_index:
                line = next(iter_lines)
                k += 1
            # Yield the lines until the mid magic comment
            while k < self.__mid_line_index:
                yield line.rstrip()
                line = next(iter_lines)
                k += 1
            # Add the mid magic comment
            yield line.rstrip()

    def close_env_lines(self) -> Iterator[str]:
        """Iterate over the script lines that close the environment."""
        with self.__script_path.open("r") as f_in_script:
            iter_lines = iter(f_in_script)
            line = next(iter_lines)
            k = 0
            # Skip before the mid magic comment
            while k < self.__mid_line_index:
                line = next(iter_lines)
                k += 1
            # Yield the lines until the end magic comment
            while k < self.__end_line_index:
                yield line.rstrip()
                line = next(iter_lines)
                k += 1
            # Add the end magic comment
            yield line.rstrip()



"""Tool abstract application module."""

# Due to typer usage:
# ruff: noqa: TC003, FBT002, FBT001

from __future__ import annotations

import datetime
import logging
from abc import ABC, abstractmethod
from collections.abc import Callable
from enum import StrEnum
from pathlib import Path
from typing import Annotated, TypeVar, cast, final

import typer

import slurmbench.experiment.checks as exp_checks
import slurmbench.experiment.file_system as exp_fs
import slurmbench.experiment.managers as exp_managers
import slurmbench.experiment.report as exp_report
import slurmbench.experiment.resume as exp_resume
import slurmbench.experiment.run as exp_run
import slurmbench.samples.status as smp_status
import slurmbench.slurm.config as slurm_cfg
from slurmbench import root_logging

from . import bash
from . import config as cfg
from . import connector as tool_connector
from . import environments as tool_env

_LOGGER = logging.getLogger(__name__)


def log_filename() -> Path:
    """Get log filename."""
    return Path(
        "pbfbench_"
        + datetime.datetime.now(tz=None).strftime("%Y-%m-%dT%H_%M_%S")  # noqa: DTZ005
        + ".log",
    )


class RichHelp(StrEnum):
    """Rich help categories."""

    MAIN_CMD = "Main commands"
    UTILS_CMD = "Utilities"


def new(
    connector_type: type[tool_connector.OnlyOptions | tool_connector.WithArguments],
) -> typer.Typer:
    """Build tool application."""
    tool_description = connector_type.description()
    app = typer.Typer(
        name=tool_description.cmd(),
        help=f"Subcommand for tool `{tool_description.name()}`",
        rich_markup_mode="rich",
    )
    #
    # Run app
    #
    run_app = Run(connector_type)
    app.command(
        name=run_app.NAME,
        help=run_app.help(),
        rich_help_panel=RichHelp.MAIN_CMD,
    )(run_app.main)
    #
    # Resume app
    #
    resume_app = Resume(connector_type)
    app.command(
        name=resume_app.NAME,
        help=resume_app.help(),
        rich_help_panel=RichHelp.MAIN_CMD,
    )(resume_app.main)
    #
    # Config app
    #
    config_app: ConfigAppWithOptions
    if connector_type is tool_connector.OnlyOptions:
        config_app = ConfigAppOnlyOptions(
            cast("type[tool_connector.OnlyOptions]", connector_type),
        )
    else:
        config_app = ConfigAppWithArguments(
            cast("type[tool_connector.WithArguments]", connector_type),
        )
    app.command(
        name=config_app.NAME,
        help=config_app.help(),
        rich_help_panel=RichHelp.UTILS_CMD,
    )(config_app.main)
    #
    # Draft tool environment script app
    #
    draft_env_sh_app = ToolEnvWrapper(connector_type)
    app.command(
        name=draft_env_sh_app.NAME,
        help=draft_env_sh_app.help(),
        rich_help_panel=RichHelp.UTILS_CMD,
    )(draft_env_sh_app.main)
    #
    # Report app
    #
    report_app = Report(connector_type)
    app.command(
        name=report_app.NAME,
        help=report_app.help(),
        rich_help_panel=RichHelp.UTILS_CMD,
    )(report_app.main)
    return app


class Arguments:
    """Tool application arguments."""

    EXP_NAME = typer.Argument(
        help="Name of the experiment",
    )

    DATA_DIR = typer.Argument(
        help="Path to the data directory (preferably absolute)",
    )


class RunArgs:
    """Run command arguments."""

    WORK_DIR = typer.Argument(
        help="Path to the working directory (preferably absolute)",
    )
    EXP_CONFIG_YAML = typer.Argument(
        help="Path to the experiment configuration YAML file (preferably absolute)",
    )


class RunOptions:
    """Run command options."""

    __RUN_CATEGORY = "Samples to run"
    __SLURM_OPTS = "SLURM configurations"

    # FEATURE Implement run options
    RUN_SUCCESS = typer.Option(
        "--success/--skip-success",
        help="Run the experiment for samples that succeeded",
        rich_help_panel=__RUN_CATEGORY,
    )
    RUN_NOT_RUN = typer.Option(
        "--not-run/--skip-not-run",
        help="Run the experiment for samples that were not run (default)",
        rich_help_panel=__RUN_CATEGORY,
    )
    RUN_MISSING_INPUTS = typer.Option(
        "--missing-inputs/--skip-missing-inputs",
        help="Run the experiment for samples that have missing inputs",
        rich_help_panel=__RUN_CATEGORY,
    )
    RUN_ERROR = typer.Option(
        "--error/--skip-error",
        help="Run the experiment for failed samples",
        rich_help_panel=__RUN_CATEGORY,
    )
    RUN_ALL = typer.Option(
        "--all",
        help="Run the experiment for all samples",
        rich_help_panel=__RUN_CATEGORY,
    )
    SLURM_OPTIONS = typer.Option(
        "--slurm-opts",
        help="SLURM options",
        rich_help_panel=__SLURM_OPTS,
    )


Connector = TypeVar(
    "Connector",
    bound=tool_connector.OnlyOptions | tool_connector.WithArguments,
)


class Run[
    CType: type[tool_connector.OnlyOptions | tool_connector.WithArguments],
]:
    """Run application."""

    NAME = "run"

    def __init__(
        self,
        tool_connector_type: CType,
    ) -> None:
        """Initialize."""
        self._tool_connector_type = tool_connector_type

    def connector_type(
        self,
    ) -> CType:
        """Get connector."""
        return self._tool_connector_type

    def help(self) -> str:
        """Get help string."""
        return (
            "Run"
            f" {self.connector_type().description().name()}"
            f" ({self.connector_type().description().topic().name()}) tool."
        )

    def main(  # noqa: PLR0913
        self,
        exp_name: Annotated[str, Arguments.EXP_NAME],
        data_dir: Annotated[Path, Arguments.DATA_DIR],
        work_dir: Annotated[Path, RunArgs.WORK_DIR],
        exp_config_yaml: Annotated[Path, RunArgs.EXP_CONFIG_YAML],
        run_success: Annotated[bool, RunOptions.RUN_SUCCESS] = False,
        run_not_run: Annotated[bool, RunOptions.RUN_NOT_RUN] = True,
        run_missing_inputs: Annotated[bool, RunOptions.RUN_MISSING_INPUTS] = False,
        run_error: Annotated[bool, RunOptions.RUN_ERROR] = False,
        run_all: Annotated[bool, RunOptions.RUN_ALL] = False,
        slurm_opts: Annotated[str | None, RunOptions.SLURM_OPTIONS] = None,
        debug: Annotated[bool, root_logging.OPT_DEBUG] = False,
    ) -> None:
        """Run tool."""
        root_logging.init_logger(
            _LOGGER,
            (
                f"Run experiment `{exp_name}`"
                f" for tool {self.connector_type().description().name()}"
                f" for topic {self.connector_type().description().topic().name()}"
            ),
            debug,
            log_file=log_filename(),
        )

        exp_manager = self._successfull_check_before_start_or_errror(
            exp_name,
            data_dir,
            work_dir,
            exp_config_yaml,
        )
        self._error_if_experiment_is_running(exp_manager)

        if slurm_opts is None:
            slurm_opts = slurm_cfg.default_slurm_options(None)

        exp_run.start_new_experiment(
            exp_manager,
            self._target_samples_to_run(
                run_success,
                run_not_run,
                run_missing_inputs,
                run_error,
                run_all,
            ),
            slurm_opts,
        )
        raise typer.Exit(0)

    def _error_if_experiment_is_running(
        self,
        exp_manager: exp_managers.WithOptions,
    ) -> None:
        """Exit with error if the experiment is already running."""
        match running_exp := exp_report.running_experiment(
            exp_manager.data_fs_manager(),
        ):
            case exp_report.RunningExperiment():
                _LOGGER.critical(
                    "The experiment is already in progress\n"
                    "* Experiment launched the: %s\n"
                    "* Working directory root path: %s\n"
                    "* SLURM job ID: %s\n"
                    "* SACCT state: %s\n",
                    running_exp.in_progress_data().date(),
                    running_exp.in_progress_data().working_directory(),
                    running_exp.in_progress_data().job_id(),
                    running_exp.sacct_state(),
                )
                _LOGGER.info("You must use the `resume` command")
                raise typer.Exit(1)

    def _successfull_check_before_start_or_errror(
        self,
        exp_name: str,
        data_dir: Path,
        work_dir: Path,
        tool_config_yaml: Path,
    ) -> exp_managers.OnlyOptions | exp_managers.WithArguments:
        #
        # Resolve absolute paths
        #
        data_dir = data_dir.resolve()
        work_dir = work_dir.resolve()
        tool_config_yaml = tool_config_yaml.resolve()

        match check_result := exp_checks.check_before_start(
            exp_name,
            data_dir,
            work_dir,
            tool_config_yaml,
            self._tool_connector_type,
        ):
            case exp_checks.RunOK():
                return check_result.exp_manager()
            case exp_checks.RunErrors():
                _LOGGER.critical("The experiment checkers found errors")
                raise typer.Exit(1)

    def _target_samples_to_run(
        self,
        run_success: bool,
        run_not_run: bool,
        run_missing_inputs: bool,
        run_error: bool,
        run_all: bool,
    ) -> Callable[[smp_status.Status], bool]:
        targets_to_run_filter: dict[smp_status.Status, bool] = dict.fromkeys(
            (
                smp_status.Success.OK,
                smp_status.Error.NOT_RUN,
                smp_status.Error.MISSING_INPUTS,
                smp_status.Error.ERROR,
            ),
            run_all,
        )
        if not run_all:
            targets_to_run_filter[smp_status.Success.OK] = run_success
            targets_to_run_filter[smp_status.Error.NOT_RUN] = run_not_run
            targets_to_run_filter[smp_status.Error.MISSING_INPUTS] = run_missing_inputs
            targets_to_run_filter[smp_status.Error.ERROR] = run_error
        return lambda status: targets_to_run_filter[status]


class Resume[
    CType: (type[tool_connector.OnlyOptions | tool_connector.WithArguments]),
]:
    """Resume application class."""

    NAME = "resume"

    def __init__(
        self,
        tool_connector_type: CType,
    ) -> None:
        """Initialize."""
        self._tool_connector_type: CType = tool_connector_type

    def connector_type(
        self,
    ) -> CType:
        """Get connector."""
        return self._tool_connector_type

    def main(
        self,
        exp_name: Annotated[str, Arguments.EXP_NAME],
        data_dir: Annotated[Path, Arguments.DATA_DIR],
        debug: Annotated[bool, root_logging.OPT_DEBUG] = False,
    ) -> None:
        """Resume the tool jobs."""
        root_logging.init_logger(
            _LOGGER,
            (
                f"Resume experiment `{exp_name}`"
                f" for tool {self.connector_type().description().name()}"
                f" for topic {self.connector_type().description().topic().name()}"
            ),
            debug,
            log_file=log_filename(),
        )

        exp_manager, array_job_id = self._retrieve_exp_manager_array_job_id(
            exp_name,
            data_dir,
        )

        exp_resume.resume(exp_manager, array_job_id)

        raise typer.Exit(0)

    def _retrieve_exp_manager_array_job_id(
        self,
        exp_name: str,
        data_dir: Path,
    ) -> tuple[exp_managers.OnlyOptions | exp_managers.WithArguments, str]:
        #
        # Resolve absolute paths
        #
        data_fs_manager = exp_fs.DataManager(
            data_dir.resolve(),
            self.connector_type().description(),
            exp_name,
        )
        running_experiment = exp_report.running_experiment(data_fs_manager)
        if running_experiment is None:
            _LOGGER.critical(
                "The experiment `%s` is not in progress for tool `%s` and topic `%s`",
                exp_name,
                self.connector_type().description().name(),
                self.connector_type().description().topic().name(),
            )
            raise typer.Exit(1)

        work_fs_manager = exp_fs.WorkManager(
            running_experiment.in_progress_data().working_directory(),
            self.connector_type().description(),
            exp_name,
        )
        connector = exp_checks.instantiate_connector(
            self.connector_type(),
            data_fs_manager.config_yaml(),
        )
        match connector:
            case tool_connector.OnlyOptions():
                return exp_managers.OnlyOptions(
                    exp_name,
                    data_fs_manager,
                    work_fs_manager,
                    connector,
                ), running_experiment.in_progress_data().job_id()
            case tool_connector.WithArguments():
                return exp_managers.WithArguments(
                    exp_name,
                    data_fs_manager,
                    work_fs_manager,
                    connector,
                ), running_experiment.in_progress_data().job_id()
            case exp_checks.RunErrors():
                _LOGGER.critical("Cannot instantiate connector")
                raise typer.Exit(1)

    def help(self) -> str:
        """Get help string."""
        return (
            "Complete pbfbench jobs for"
            f" {self.connector_type().description().name()}"
            f" ({self.connector_type().description().topic().name()}) tool."
        )


class ConfigAppWithOptions[
    Connector: tool_connector.WithOptions,
    Config: cfg.WithOptions,
](ABC):
    """Config application base class."""

    NAME = "config"

    def __init__(self, connector_type: type[Connector]) -> None:
        """Initialize."""
        self._connector_type = connector_type

    def connector_type(self) -> type[Connector]:
        """Get connector."""
        return self._connector_type

    def help(self) -> str:
        """Get help string."""
        return (
            "Get draft "
            + self._connector_type.description().name()
            + " tool configuration."
        )

    def main(
        self,
        config_exp_yaml: Annotated[Path, RunArgs.EXP_CONFIG_YAML],
        debug: Annotated[bool, root_logging.OPT_DEBUG] = False,
    ) -> None:
        """Get draft config."""
        root_logging.init_logger(
            _LOGGER,
            "Generating a configuration file draft"
            f" for topic: {self._connector_type.description().topic().name()}"
            f" tool: {self._connector_type.description().name()}",
            debug,
        )

        config = self._fake_config()

        config.to_yaml(config_exp_yaml)
        _LOGGER.info("Tool configuration written to %s", config_exp_yaml)
        raise typer.Exit(0)

    @abstractmethod
    def _fake_config(self) -> Config:
        """Create fake tool config."""
        raise NotImplementedError

    def _fake_options_config(self) -> cfg.StringOpts:
        return cfg.StringOpts(
            ("--options1=value1", "--options2=value2"),
        )


@final
class ConfigAppOnlyOptions(
    ConfigAppWithOptions[tool_connector.OnlyOptions, cfg.OnlyOptions],
):
    """Config application for tools with only options."""

    def _fake_config(self) -> cfg.OnlyOptions:
        return self._connector_type.config_type()(self._fake_options_config())


@final
class ConfigAppWithArguments(
    ConfigAppWithOptions[tool_connector.WithArguments, cfg.WithArguments],
):
    """Config application for tools with arguments."""

    def _fake_config(self) -> cfg.WithArguments:
        return self._connector_type.config_type()(
            self._fake_arguments_config(),
            self._fake_options_config(),
        )

    def _fake_arguments_config(self) -> cfg.Arguments:
        """Create arguments."""
        args: dict[str, cfg.Arg] = {}
        for arg_type in self._connector_type.arguments_type().arg_types():
            tool_choice_str = " | ".join(map(str, arg_type.valid_tools()))
            if not tool_choice_str:
                tool_choice_str = "ERROR: no tool implements this argument"
            args[str(arg_type.name())] = cfg.Arg(
                tool_choice_str,
                "$input_experiment_name",
            )
        return cfg.Arguments(args)


class ToolEnvWrapperOpts:
    """Tool environnment wrapper options."""

    ERASE = typer.Option(
        "--erase/--do-not-erase",
        help="Erase existing tool environnment wrapper script",
    )


class ToolEnvWrapper[CType: type[tool_connector.WithOptions]]:
    """Draft tool environnment wrapper script."""

    NAME = "draft-env"

    def __init__(self, tool_connector_type: CType) -> None:
        self._tool_connector_type = tool_connector_type

    def main(
        self,
        data_dir: Annotated[Path, Arguments.DATA_DIR],
        erase: Annotated[bool, ToolEnvWrapperOpts.ERASE] = False,
        debug: Annotated[bool, root_logging.OPT_DEBUG] = False,
    ) -> None:
        """Generate environnment wrapper script."""
        root_logging.init_logger(
            _LOGGER,
            (
                f"Generating an environnment wrapper script"
                f" for tool {self._tool_connector_type.description().name()}"
                f" for topic {self._tool_connector_type.description().topic().name()}"
            ),
            debug,
        )
        data_exp_manager = exp_fs.DataManager(
            data_dir.resolve(),
            self._tool_connector_type.description(),
            "fake_exp",
        )
        core_command_sh_path = bash.WithOptions.core_command_sh_path(
            self._tool_connector_type.parent_dir_where_defined(),
        )
        env_wrapper_sh_path = data_exp_manager.tool_env_script_sh()
        if env_wrapper_sh_path.exists() and not erase:
            _LOGGER.critical(
                "The environnment wrapper script %s already exists",
                env_wrapper_sh_path,
            )
            raise typer.Exit(1)

        if not data_exp_manager.tool_dir().exists():
            _LOGGER.info(
                "Create non-existing tool directory %s",
                data_exp_manager.tool_dir(),
            )
            data_exp_manager.tool_dir().mkdir(parents=True)

        to_do_lines: list[str] = []
        with core_command_sh_path.open() as f_in:
            in_block_comment = True
            iter_lines = iter(f_in)
            while in_block_comment:
                line = next(iter_lines, None)
                if line is None or not line.startswith("#"):
                    in_block_comment = False
                else:
                    to_do_lines.append(line.rstrip())

        with env_wrapper_sh_path.open("w") as f_out:
            f_out.write("# Tips from the tool core commands:\n")
            f_out.write("\n".join(to_do_lines))
            if to_do_lines:
                f_out.write("\n")
            f_out.write("\n")
            f_out.write(tool_env.ScriptWrapper.BEGIN_ENV_MAGIC_COMMENT + " ===\n")
            f_out.write("\n")
            f_out.write("# Commands to initialize the environment\n")
            f_out.write("\n")
            f_out.write(tool_env.ScriptWrapper.MID_ENV_MAGIC_COMMENT + " ---\n")
            f_out.write("\n")
            f_out.write("# Commands to close the environment\n")
            f_out.write("\n")
            f_out.write(tool_env.ScriptWrapper.END_ENV_MAGIC_COMMENT + " ===\n")

        _LOGGER.info(
            "Tool environnment wrapper draft script written to %s",
            env_wrapper_sh_path,
        )
        raise typer.Exit(0)

    def help(self) -> str:
        """Get help string."""
        return (
            "Generate a draft environnment wrapper script"
            f" for tool {self._tool_connector_type.description().name()}"
            f" for topic {self._tool_connector_type.description().topic().name()}"
        )


class ReportOpts:
    """Report command options."""

    YAML_FILE = typer.Option(
        "--yaml",
        "-o",
        help="YAML output file",
    )


class Report[CType: type[tool_connector.WithOptions]]:
    """Report command."""

    NAME = "report"

    def __init__(self, tool_connector_type: CType) -> None:
        self._tool_connector_type = tool_connector_type

    def main(
        self,
        exp_name: Annotated[str, Arguments.EXP_NAME],
        data_dir: Annotated[Path, Arguments.DATA_DIR],
        yaml_file: Annotated[Path | None, ReportOpts.YAML_FILE] = None,
        debug: Annotated[bool, root_logging.OPT_DEBUG] = False,
    ) -> None:
        """Generate report."""
        root_logging.init_logger(
            _LOGGER,
            (
                f"Generating a report"
                f" for tool {self._tool_connector_type.description().name()}"
                f" for topic {self._tool_connector_type.description().topic().name()}"
            ),
            debug,
        )
        data_exp_manager = exp_fs.DataManager(
            data_dir.resolve(),
            self._tool_connector_type.description(),
            exp_name,
        )
        if not data_exp_manager.exp_dir().exists():
            _LOGGER.critical(
                "There is no experiment nammed `%s` for tool `%s` and topic `%s`",
                data_exp_manager.experiment_name(),
                self._tool_connector_type.description().name(),
                self._tool_connector_type.description().topic().name(),
            )
            raise typer.Exit(1)

        report = exp_report.Report.from_data_exp_fs_manager(data_exp_manager)

        _LOGGER.info("Report:\n%s", exp_report.report_to_string(report))
        if yaml_file is not None:
            _LOGGER.info("Writing report to `%s`", yaml_file)
            report.to_yaml(yaml_file)
        raise typer.Exit(0)

    def help(self) -> str:
        """Get help string."""
        return (
            "Generate a report"
            f" for tool {self._tool_connector_type.description().name()}"
            f" for topic {self._tool_connector_type.description().topic().name()}"
        )



"""Tool config YAML interface."""

from __future__ import annotations

from typing import TYPE_CHECKING, Self, final

from slurmbench.yaml_interface import YAMLInterface

if TYPE_CHECKING:
    from collections.abc import Iterable, Iterator

    import yaml  # type: ignore[import-untyped]

from abc import ABC


@final
class Arg(YAMLInterface):
    """Tool config argument."""

    @classmethod
    def from_yaml_load(cls, pyyaml_obj: list[str]) -> Self:
        """Convert dict to object."""
        return cls(*pyyaml_obj)

    def __init__(self, tool_name: str, exp_name: str) -> None:
        self._tool_name = tool_name
        self._exp_name = exp_name

    def tool_name(self) -> str:
        """Get tool name."""
        return self._tool_name

    def exp_name(self) -> str:
        """Get experiment name."""
        return self._exp_name

    def to_yaml_dump(self) -> list[str]:
        """Convert to yaml dump."""
        return [self._tool_name, self._exp_name]


@final
class Arguments(YAMLInterface):
    """Tool config arguments."""

    @classmethod
    def from_yaml_load(cls, pyyaml_obj: dict[str, list[str]]) -> Self:
        """Convert dict to object."""
        return cls({name: Arg.from_yaml_load(arg) for name, arg in pyyaml_obj.items()})

    def __init__(self, arguments: dict[str, Arg]) -> None:
        self._arguments = arguments

    def arguments(self) -> dict[str, Arg]:
        """Get arguments."""
        return self._arguments

    def __getitem__(self, name: str) -> Arg:
        """Get argument."""
        return self._arguments[name]

    def to_yaml_dump(self) -> dict[str, list[str]]:
        """Convert to yaml dump."""
        return {name: arg.to_yaml_dump() for name, arg in self._arguments.items()}


@final
class StringOpts(YAMLInterface):
    """String options.

    When the options are regular short/long options.
    """

    @classmethod
    def from_yaml_load(cls, obj_list: list[str]) -> Self:
        """Convert dict to object."""
        return cls(obj_list)

    def __init__(self, options: Iterable[str]) -> None:
        self.__options = list(options)

    def __bool__(self) -> bool:
        """Check if options are not empty."""
        return len(self.__options) > 0

    def __len__(self) -> int:
        """Get options length."""
        return len(self.__options)

    def __iter__(self) -> Iterator[str]:
        """Iterate options."""
        return iter(self.__options)

    def to_yaml_dump(self) -> list[str]:
        """Convert to dict."""
        return self.__options


class WithOptions(YAMLInterface, ABC):
    """Tool config with options."""

    KEY_OPTIONS = "options"

    @classmethod
    def _get_options_from_yaml_load(
        cls,
        yaml_obj: dict[str, yaml.YAMLObject],
    ) -> StringOpts:
        return StringOpts.from_yaml_load(yaml_obj.get(cls.KEY_OPTIONS, []))

    def __init__(self, options: StringOpts) -> None:
        self._options = options

    def options(self) -> StringOpts:
        """Get options."""
        return self._options

    def _options_to_yaml_dump(self) -> dict[str, list[str]]:
        if not self._options:
            return {}
        return {self.KEY_OPTIONS: self._options.to_yaml_dump()}

    def is_same(self, other: Self) -> bool:
        """Check if configs are the same."""
        return self.to_yaml_dump() == other.to_yaml_dump()


@final
class OnlyOptions(WithOptions):
    """Tool config without arguments."""

    @classmethod
    def from_yaml_load(cls, obj_dict: dict[str, yaml.YAMLObject]) -> Self:
        """Convert dict to object."""
        return cls(cls._get_options_from_yaml_load(obj_dict))

    def to_yaml_dump(self) -> dict[str, yaml.YAMLObject]:
        """Convert to dict."""
        return self._options_to_yaml_dump()


@final
class WithArguments(WithOptions):
    """Tool config with arguments."""

    KEY_ARGUMENTS = "arguments"

    @classmethod
    def from_yaml_load(cls, obj_dict: dict[str, yaml.YAMLObject]) -> Self:
        """Convert dict to object."""
        return cls(
            Arguments.from_yaml_load(obj_dict[cls.KEY_ARGUMENTS]),
            cls._get_options_from_yaml_load(obj_dict),
        )

    def __init__(self, arguments: Arguments, options: StringOpts) -> None:
        """Initialize."""
        super().__init__(options)
        self.__arguments = arguments

    def arguments(self) -> Arguments:
        """Get arguments."""
        return self.__arguments

    def to_yaml_dump(self) -> dict[str, yaml.YAMLObject]:
        """Convert to dict."""
        return {
            self.KEY_ARGUMENTS: self.__arguments.to_yaml_dump(),
            **self._options_to_yaml_dump(),
        }



"""Experiment stats module."""

from __future__ import annotations

from typing import Self, final

import slurmbench.samples.status as smp_status
from slurmbench.yaml_interface import YAMLInterface


@final
class Samples(YAMLInterface):
    """Stats of samples for an experiment."""

    KEY_TOTAL_NUMBER_OF_SAMPLES = "total_number_of_samples"
    KEY_NUMBER_OF_SUCCESSFUL_SAMPLES = "number_of_successful_samples"
    KEY_NUMBER_OF_SAMPLES_WITH_MISSING_INPUTS = "number_of_samples_with_missing_inputs"
    KEY_NUMBER_OF_FAILED_SAMPLES = "number_of_failed_samples"
    KEY_NUMBER_OF_NOT_RUN_SAMPLES = "number_of_not_run_samples"

    @classmethod
    def from_status_map(cls, status_map: smp_status.StatusMap[int]) -> Self:
        """Convert status map to stats."""
        return cls(
            status_map[smp_status.Success.OK],
            status_map[smp_status.Error.MISSING_INPUTS],
            status_map[smp_status.Error.ERROR],
            status_map[smp_status.Error.NOT_RUN],
        )

    @classmethod
    def from_yaml_load(cls, pyyaml_obj: dict[str, int]) -> Self:
        """Convert dict to object."""
        return cls(
            int(pyyaml_obj[cls.KEY_NUMBER_OF_SUCCESSFUL_SAMPLES]),
            int(pyyaml_obj[cls.KEY_NUMBER_OF_SAMPLES_WITH_MISSING_INPUTS]),
            int(pyyaml_obj[cls.KEY_NUMBER_OF_FAILED_SAMPLES]),
            int(pyyaml_obj[cls.KEY_NUMBER_OF_NOT_RUN_SAMPLES]),
        )

    def __init__(
        self,
        number_of_successful_samples: int,
        number_of_samples_with_missing_inputs: int,
        number_of_failed_samples: int,
        number_of_not_run_samples: int,
    ) -> None:
        self._number_of_successful_samples = number_of_successful_samples
        self._number_of_samples_with_missing_inputs = (
            number_of_samples_with_missing_inputs
        )
        self._number_of_failed_samples = number_of_failed_samples
        self._number_of_not_run_samples = number_of_not_run_samples

    def total_number_of_samples(self) -> int:
        """Get total number of samples."""
        return (
            self._number_of_successful_samples
            + self._number_of_samples_with_missing_inputs
            + self._number_of_failed_samples
            + self._number_of_not_run_samples
        )

    def number_of_successful_samples(self) -> int:
        """Get number of successful samples."""
        return self._number_of_successful_samples

    def number_of_samples_with_missing_inputs(self) -> int:
        """Get number of samples with missing inputs."""
        return self._number_of_samples_with_missing_inputs

    def number_of_failed_samples(self) -> int:
        """Get number of failed samples."""
        return self._number_of_failed_samples

    def number_of_not_run_samples(self) -> int:
        """Get number of not run samples."""
        return self._number_of_not_run_samples

    def to_yaml_dump(self) -> dict[str, int]:
        """Convert to dict."""
        return {
            self.KEY_TOTAL_NUMBER_OF_SAMPLES: self.total_number_of_samples(),
            self.KEY_NUMBER_OF_SUCCESSFUL_SAMPLES: self._number_of_successful_samples,
            self.KEY_NUMBER_OF_SAMPLES_WITH_MISSING_INPUTS: (
                self._number_of_samples_with_missing_inputs
            ),
            self.KEY_NUMBER_OF_FAILED_SAMPLES: self._number_of_failed_samples,
            self.KEY_NUMBER_OF_NOT_RUN_SAMPLES: self._number_of_not_run_samples,
        }



"""Experiment run module."""

from __future__ import annotations

import logging
import shutil
from typing import TYPE_CHECKING, cast

import typer

import slurmbench.samples.items as smp
import slurmbench.samples.missing_inputs as smp_miss_in
import slurmbench.samples.status as smp_status
import slurmbench.tool.connector as tool_connector
import slurmbench.topic.results as topic_res

from . import checks as exp_checks
from . import errors as exp_errors
from . import file_system as exp_fs
from . import in_progress, monitors, resolve
from . import iter as exp_iter
from . import managers as exp_managers
from .bash import create as exp_bash_create
from .slurm import run as exp_slurm_run

if TYPE_CHECKING:
    from collections.abc import Callable

_LOGGER = logging.getLogger(__name__)


def start_new_experiment(
    exp_manager: exp_managers.OnlyOptions | exp_managers.WithArguments,
    target_samples_filter: Callable[[smp_status.Status], bool],
    slurm_opts: str,
) -> None:
    """Run the experiment."""
    # REFACTOR use markdown print and do better app prints

    _LOGGER.info(
        "Running experiment `%s` with tool `%s` for the topic `%s`.",
        exp_manager.exp_name(),
        exp_manager.tool_connector().description().name(),
        exp_manager.tool_connector().description().topic().name(),
    )

    _reset_experiment_working_directory(exp_manager)

    _first_experiment_run_or_check_same_config(exp_manager)

    samples_to_run = _get_samples_to_run(exp_manager, target_samples_filter)
    if not samples_to_run:
        _LOGGER.info("No samples to run")
        return

    if isinstance(exp_manager, exp_managers.WithArguments):
        samples_to_run = _manage_inputs(exp_manager, samples_to_run)

    if not samples_to_run:
        _LOGGER.info("No samples to run")
        resolve.conclude_experiment(exp_manager)
        return

    _LOGGER.info("Number of samples sent to sbatch: %d", len(samples_to_run))

    sh_manager = exp_bash_create.run_scripts(
        exp_manager,
        samples_to_run,
        slurm_opts,
    )

    exp_slurm_run.run(sh_manager)

    job_id = in_progress.write_in_progress_metadata(exp_manager, sh_manager)

    resolve.resolve_running_samples(exp_manager, samples_to_run, job_id)


def _reset_experiment_working_directory(exp_manager: exp_managers.WithOptions) -> None:
    """Reset experiment working directory."""
    if exp_manager.work_fs_manager().exp_dir().exists():
        _LOGGER.info(
            "Removing experiment working directory: %s",
            exp_manager.work_fs_manager().exp_dir(),
        )
        shutil.rmtree(exp_manager.work_fs_manager().exp_dir())
    exp_manager.work_fs_manager().exp_dir().mkdir(parents=True)


def _first_experiment_run_or_check_same_config(
    exp_manager: exp_managers.OnlyOptions | exp_managers.WithArguments,
) -> None:
    """Check if the experiment has been run before and check the configs."""
    if exp_manager.data_fs_manager().history_yaml().exists():
        match exp_checks.compare_config_vs_config_in_data(
            exp_manager.tool_connector(),
            exp_manager.data_fs_manager().config_yaml(),  # must also exist
        ):
            case exp_checks.DifferentExperimentConfigs():
                raise typer.Exit(1)
    else:
        exp_manager.data_fs_manager().exp_dir().mkdir(parents=True, exist_ok=True)
        # Copy config files
        exp_manager.tool_connector().to_config().to_yaml(
            exp_manager.data_fs_manager().config_yaml(),
        )
        exp_manager.tool_connector().to_config().to_yaml(
            exp_manager.work_fs_manager().config_yaml(),
        )


def _get_samples_to_run(
    exp_manager: exp_managers.OnlyOptions | exp_managers.WithArguments,
    target_samples_filter: Callable[[smp_status.Status], bool],
) -> list[smp.RowNumbered]:
    """Get samples to run."""
    samples_to_run = [
        row_numbered_item
        for row_numbered_item, status in exp_iter.samples_with_status(
            exp_manager.data_fs_manager(),
        )
        if target_samples_filter(status)
    ]
    _LOGGER.info("Number of samples to run: %d", len(samples_to_run))

    _remove_from_previous_error_list_samples_to_run(exp_manager, samples_to_run)

    monitors.write_unresolved_samples(exp_manager.work_fs_manager(), samples_to_run)

    return samples_to_run


def _remove_from_previous_error_list_samples_to_run(
    exp_manager: exp_managers.OnlyOptions | exp_managers.WithArguments,
    samples_to_run: list[smp.RowNumbered],
) -> None:
    """Remove from previous error list samples to run."""
    if not exp_manager.data_fs_manager().errors_tsv().exists():
        return

    sample_ids_to_run = {
        row_numbered_item.uid() for row_numbered_item in samples_to_run
    }

    with exp_errors.ErrorsTSVReader.open(
        exp_manager.data_fs_manager().errors_tsv(),
    ) as errors_tsv_reader:
        to_keep_in_list = [
            sample_error
            for sample_error in errors_tsv_reader
            if sample_error.sample_uid() not in sample_ids_to_run
        ]
    with exp_errors.ErrorsTSVWriter.open(
        exp_manager.data_fs_manager().errors_tsv(),
        "w",
    ) as errors_tsv_writer:
        errors_tsv_writer.write_error_samples(to_keep_in_list)


def _manage_inputs(
    exp_manager: exp_managers.WithArguments,
    samples_to_run: list[smp.RowNumbered],
) -> list[smp.RowNumbered]:
    """Manage inputs."""
    _format_inputs(exp_manager, samples_to_run)
    return _filter_missing_inputs(exp_manager, samples_to_run)


def _format_inputs[N: tool_connector.Names](
    exp_manager: exp_managers.WithArguments[N],
    samples_to_run: list[smp.RowNumbered],
) -> None:
    tool_inputs_to_fmt: list[tuple[exp_fs.DataManager, topic_res.ConvertFn]] = []
    missing_convert_fn = False
    for _, arg in exp_manager.tool_connector().arguments():
        result_visitor = arg.result_visitor()
        if result_visitor is topic_res.FormattedVisitor:
            result_visitor = cast(
                "type[topic_res.FormattedVisitor]",
                result_visitor,
            )
            convert_fn_or_err = result_visitor.convert_fn(arg.tool())
            if isinstance(convert_fn_or_err, topic_res.Error):
                _LOGGER.critical("%s", convert_fn_or_err)
                missing_convert_fn = True
                continue

            in_data_exp_fs_manager = exp_fs.DataManager(
                exp_manager.data_fs_manager().root_dir(),
                arg.tool().to_description(),
                arg.exp_name(),
            )

            tool_inputs_to_fmt.append((in_data_exp_fs_manager, convert_fn_or_err))

    if missing_convert_fn:
        return

    for row_numbered_sample in samples_to_run:
        for in_data_exp_fs_manager, convert_fn in tool_inputs_to_fmt:
            # OPTIMIZE do not redo format if exists?
            try:
                convert_fn(in_data_exp_fs_manager, row_numbered_sample)
            except Exception:
                _LOGGER.exception(
                    "Error formatting sample %s",
                    row_numbered_sample.uid(),
                )


def _filter_missing_inputs[N: tool_connector.Names](
    exp_manager: exp_managers.WithArguments[N],
    samples_to_run: list[smp.RowNumbered],
) -> list[smp.RowNumbered]:
    """Filter missing inputs."""
    if not samples_to_run:
        return []

    samples_without_missing_inputs: list[smp.RowNumbered] = []
    samples_with_missing_inputs: list[smp.RowNumbered] = []

    tool_inputs = dict(
        exp_manager.tool_connector().arguments().results(exp_manager.data_fs_manager()),
    )

    for row_numbered_sample in samples_to_run:
        sample_missing_inputs = smp_miss_in.for_sample(
            tool_inputs,
            row_numbered_sample,
        )

        if sample_missing_inputs:
            samples_with_missing_inputs.append(row_numbered_sample)
            smp_miss_in.write_sample_missing_inputs(
                exp_manager,
                row_numbered_sample,
                sample_missing_inputs,
            )
        else:
            samples_without_missing_inputs.append(row_numbered_sample)

    if samples_with_missing_inputs:
        _LOGGER.error(
            "Samples with missing inputs: %d",
            len(samples_with_missing_inputs),
        )
        exp_errors.write_missing_inputs(
            exp_manager.data_fs_manager(),
            samples_with_missing_inputs,
        )
        monitors.update_samples_resolution_status(
            exp_manager.work_fs_manager(),
            ((s, smp_status.Error.MISSING_INPUTS) for s in samples_with_missing_inputs),
        )

    return samples_without_missing_inputs



"""Tools script logics."""

from __future__ import annotations

import logging
import shutil
import stat
from itertools import chain
from pathlib import Path
from typing import TYPE_CHECKING, assert_never

import typer

import slurmbench.bash.items as bash_items
import slurmbench.experiment.file_system as exp_fs
import slurmbench.experiment.managers as exp_managers
import slurmbench.experiment.slurm.status as exp_slurm_status
import slurmbench.samples.items as smp
import slurmbench.slurm.bash as slurm_bash
import slurmbench.tool.bash as tool_bash
import slurmbench.tool.environments as tools_envs

from . import items as exp_bash_items
from . import manager

if TYPE_CHECKING:
    from collections.abc import Iterable, Iterator

_LOGGER = logging.getLogger(__name__)


def run_scripts(
    exp_manager: exp_managers.OnlyOptions | exp_managers.WithArguments,
    samples_to_run: Iterable[smp.RowNumbered],
    slurm_opts: str,
) -> manager.Manager:
    """Create the run script."""
    sh_manager = manager.Manager.from_exp_fs_manager(exp_manager)
    sh_manager.data_sh_fs_manager().scripts_dir().mkdir(parents=True, exist_ok=True)
    sh_manager.work_sh_fs_manager().scripts_dir().mkdir(parents=True, exist_ok=True)

    exp_manager.work_fs_manager().slurm_log_fs_manager().log_dir().mkdir(
        parents=True,
        exist_ok=True,
    )

    tool_commands = exp_manager.tool_connector().sh_commands(
        exp_manager.data_fs_manager(),
        exp_manager.work_fs_manager(),
    )
    tool_bash_env_wrapper = tools_envs.ScriptWrapper.new(
        exp_manager.data_fs_manager().tool_env_script_sh(),
    )
    if isinstance(tool_bash_env_wrapper, tools_envs.MissingMagicCommentError):
        _LOGGER.error(tool_bash_env_wrapper)
        _LOGGER.critical("Cannot parse the tool environment wrapper script")
        raise typer.Exit(1)  # REFACTOR should use contexts instead of existing locally

    for script_path in (
        _init_env_script(sh_manager, tool_bash_env_wrapper),
        _command_script(sh_manager, tool_commands),
        _close_env_script(sh_manager, tool_bash_env_wrapper),
        _sbatch_script(exp_manager, sh_manager, samples_to_run, slurm_opts),
    ):
        _add_x_permissions(script_path)
        shutil.copy(script_path, sh_manager.data_sh_fs_manager().scripts_dir())

    return sh_manager


def _init_env_script(
    sh_manager: manager.Manager,
    tool_bash_env_wrapper: tools_envs.ScriptWrapper,
) -> Path:
    """Write the init environment script."""
    script_path = sh_manager.work_sh_fs_manager().step_script(
        exp_bash_items.Steps.INIT_ENV,
    )
    with script_path.open("w") as f_out:
        for line in tool_bash_env_wrapper.init_env_lines():
            f_out.write(line + "\n")
    return script_path


def _command_script(
    sh_manager: manager.Manager,
    tool_cmd: tool_bash.WithOptions,
) -> Path:
    """Write the command script (which `srun` will call)."""
    script_path = sh_manager.work_sh_fs_manager().step_script(
        exp_bash_items.Steps.COMMAND,
    )
    with script_path.open("w") as command_out:
        for line in _tool_command_script_lines(tool_cmd):
            command_out.write(line + "\n")
    return script_path


def _tool_command_script_lines(
    tool_cmd: tool_bash.WithOptions,
) -> Iterator[str]:
    """Return command lines."""
    yield bash_items.BASH_SHEBANG
    yield ""
    yield "set -e"  # exit error at the first command failing
    yield ""
    yield from tool_cmd.commands()


def _close_env_script(
    sh_manager: manager.Manager,
    tool_bash_env_wrapper: tools_envs.ScriptWrapper,
) -> Path:
    """Write the close environment script."""
    script_path = sh_manager.work_sh_fs_manager().step_script(
        exp_bash_items.Steps.CLOSE_ENV,
    )
    with script_path.open("w") as f_out:
        for line in tool_bash_env_wrapper.close_env_lines():
            f_out.write(line + "\n")
    return script_path


def _add_x_permissions(cmd_sh_path: Path) -> None:
    """Chmod +x the subscript (called by `srun`) for everyone."""
    st = cmd_sh_path.stat()
    cmd_sh_path.chmod(st.st_mode | stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH)


def _sbatch_script(
    exp_manager: exp_managers.OnlyOptions | exp_managers.WithArguments,
    sh_manager: manager.Manager,
    samples_to_run: Iterable[smp.RowNumbered],
    slurm_opts: str,
) -> Path:
    """Write the sbatch script."""
    sbatch_script = sh_manager.work_sh_fs_manager().sbatch_script()
    with sbatch_script.open("w") as sbatch_out:
        for line in SbatchLinesBuilder.lines(
            slurm_opts,
            sh_manager,
            exp_manager.work_fs_manager(),
            samples_to_run,
        ):
            sbatch_out.write(line + "\n")
    return sbatch_script


class StepLinesBuilder:
    """Sbatch step lines builder."""

    @classmethod
    def set_step_ok_file(
        cls,
        work_exp_fs_manager: exp_fs.WorkManager,
        step: exp_bash_items.Steps,
    ) -> Iterator[str]:
        """Set step OK file variable."""
        yield bash_items.Variable(
            f"{step}_STEP_OK_FILE",
        ).set(
            bash_items.path_to_str(
                work_exp_fs_manager.slurm_log_fs_manager().script_step_status_file(
                    slurm_bash.SLURM_JOB_ID_FROM_VARS,
                    step,
                    exp_slurm_status.ScriptSteps.OK,
                ),
            ),
        )

    @classmethod
    def step_error_file(
        cls,
        work_exp_fs_manager: exp_fs.WorkManager,
        step: exp_bash_items.Steps,
    ) -> Iterator[str]:
        """Set step error file variable."""
        yield bash_items.Variable(
            f"{step}_STEP_ERROR_FILE",
        ).set(
            bash_items.path_to_str(
                work_exp_fs_manager.slurm_log_fs_manager().script_step_status_file(
                    slurm_bash.SLURM_JOB_ID_FROM_VARS,
                    step,
                    exp_slurm_status.ScriptSteps.ERROR,
                ),
            ),
        )

    @classmethod
    def set_script_path(
        cls,
        sh_manager: manager.Manager,
        step: exp_bash_items.Steps,
    ) -> Iterator[str]:
        """Set script path variable."""
        yield bash_items.Variable(f"{step}_SCRIPT").set(
            bash_items.path_to_str(
                sh_manager.work_sh_fs_manager().step_script(step),
            ),
        )


class SbatchLinesBuilder:
    """Sbatch lines builder."""

    TEMPLATE_SCRIPT = Path(__file__).parent / "sbatch_template.sh"

    PBFBENCH_DO_PREFIX = "# PBFBENCH_DO:"

    @classmethod
    def lines(
        cls,
        slurm_opts: str,
        sh_manager: manager.Manager,
        work_exp_fs_manager: exp_fs.WorkManager,
        samples_to_run: Iterable[smp.RowNumbered],
    ) -> Iterator[str]:
        """Return sbatch lines."""
        with cls.TEMPLATE_SCRIPT.open("r") as template_in:
            for line in template_in:
                pbfbench_do = cls._pbfbench_do(line)
                if pbfbench_do is None:
                    yield line.rstrip()
                else:
                    yield from cls._pbfbench_do_lines(
                        line,
                        pbfbench_do,
                        slurm_opts,
                        sh_manager,
                        work_exp_fs_manager,
                        samples_to_run,
                    )

    @classmethod
    def _pbfbench_do(cls, line: str) -> exp_bash_items.PbfbenchDo | None:
        """Return the PBFBENCH_DO comment if it is, otherwise None."""
        if line.startswith(cls.PBFBENCH_DO_PREFIX):
            pbfbench_do_str = line[len(cls.PBFBENCH_DO_PREFIX) :].rstrip().split(":")[0]
            return exp_bash_items.PbfbenchDo(pbfbench_do_str)
        return None

    @classmethod
    def _script_step(cls, line: str) -> exp_bash_items.Steps:
        """Return the script step if it is, otherwise None."""
        step_str = line[len(cls.PBFBENCH_DO_PREFIX) :].rstrip().split(":")[1]
        return exp_bash_items.Steps(step_str)

    @classmethod
    def _pbfbench_do_lines(  # noqa: PLR0913
        cls,
        line: str,
        pbfbench_do: exp_bash_items.PbfbenchDo,
        slurm_opts: str,
        sh_manager: manager.Manager,
        work_exp_fs_manager: exp_fs.WorkManager,
        samples_to_run: Iterable[smp.RowNumbered],
    ) -> Iterator[str]:
        """Return the lines for the PBFBENCH_DO comment."""
        match pbfbench_do:
            case exp_bash_items.PbfbenchDo.SBATCH_COMMENTS:
                return slurm_bash.SbatchCommentLinesBuilder.lines(
                    slurm_opts,
                    (sample.to_base_one() for sample in samples_to_run),
                    work_exp_fs_manager,
                )
            case exp_bash_items.PbfbenchDo.ARRAY_JOB_ID_FILE:
                return iter(
                    (
                        bash_items.Variable(pbfbench_do.value).set(
                            bash_items.path_to_str(
                                work_exp_fs_manager.slurm_log_fs_manager()
                                .job_id_file_manager()
                                .path(),
                            ),
                        ),
                    ),
                )
            case exp_bash_items.PbfbenchDo.STEP:
                step = cls._script_step(line)
                return chain(
                    StepLinesBuilder.set_script_path(sh_manager, step),
                    StepLinesBuilder.set_step_ok_file(work_exp_fs_manager, step),
                    StepLinesBuilder.step_error_file(work_exp_fs_manager, step),
                )
        assert_never(pbfbench_do)



"""Topics visitor prelude."""

# Due to typer usage:
# ruff: noqa: PLC0414

from slurmbench.topic.description import Description as Description
from slurmbench.topic.visitor import Tools as Tools
from slurmbench.topics.visitor import Topics as Topics



"""Abstract tools results items module."""

from abc import ABC, abstractmethod
from collections.abc import Callable

import slurmbench.experiment.file_system as exp_fs
import slurmbench.samples.items as smp
import slurmbench.tool.results as tool_res

from . import visitor


class Error:
    """Error to get the result."""

    def __init__(self, msg: str) -> None:
        """Initialize."""
        self.__msg = msg

    def __str__(self) -> str:
        """Get the error message."""
        return self.__msg


# Separate the visitor from the result class
# because some result are really specific, other are conventionnal.
# Only conventionnal results have a visitor.
class Visitor[T: visitor.Tools, R: tool_res.Result](ABC):
    """Abstract result visitor."""

    @classmethod
    @abstractmethod
    def result_builder(cls) -> type[R]:
        """Get result builder."""
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def result_builder_from_tool(cls, tool: T) -> Error | type[R]:
        """Get result builder."""
        raise NotImplementedError

    @classmethod
    def tool_gives_the_result(cls, tool: T) -> bool:
        """Check if the tool gives the result."""
        return not isinstance(cls.result_builder_from_tool(tool), Error)


class OriginalVisitor[T: visitor.Tools, OriginalResult: tool_res.Original](
    Visitor[T, OriginalResult],
):
    """Original result visitor."""


type ConvertFn[FormattedResult: tool_res.Formatted] = Callable[
    [exp_fs.DataManager, smp.RowNumbered],
    FormattedResult,
]


class FormattedVisitor[T: visitor.Tools, FormattedResult: tool_res.Formatted](
    Visitor[T, FormattedResult],
):
    """Formatted result visitor."""

    @classmethod
    @abstractmethod
    def convert_fn(
        cls,
        tool: T,
    ) -> Error | ConvertFn[FormattedResult]:
        """Get convert function."""
        raise NotImplementedError

    @classmethod
    def result_builder_from_tool(
        cls,
        tool: T,
    ) -> Error | type[FormattedResult]:
        """Get result builder."""
        match convert_fn_or_err := cls.convert_fn(tool):
            case Error():
                return convert_fn_or_err
            case _:
                return cls.result_builder()



"""Experiment iter module."""

from __future__ import annotations

from typing import TYPE_CHECKING

import slurmbench.samples.file_system as smp_fs
import slurmbench.samples.items as smp
import slurmbench.samples.status as smp_status

if TYPE_CHECKING:
    from collections.abc import Iterator

    from . import file_system as exp_fs


def samples_with_status(
    data_exp_fs_manager: exp_fs.DataManager,
) -> Iterator[tuple[smp.RowNumbered, smp_status.Status]]:
    """Get samples with their status."""
    with smp_fs.TSVReader.open(data_exp_fs_manager.samples_tsv()) as smp_tsv_in:
        for row_numbered_sample in smp_tsv_in:
            yield (
                row_numbered_sample,
                smp_status.get_status(
                    data_exp_fs_manager.sample_fs_manager(row_numbered_sample),
                ),
            )



"""Bash shell items."""

from pathlib import Path

BASH_SHEBANG = "#!/bin/bash"


class Variable:
    """A bash variable."""

    def __init__(self, name: str) -> None:
        self.__name = name

    def name(self) -> str:
        """Get name."""
        return self.__name

    def eval(self) -> str:
        """Get evaluated variable."""
        return f"${{{self.__name}}}"

    def set(self, value: str) -> str:
        """Set variable."""
        return f"{self.__name}={value}"

    def set_path(self, path: Path | str) -> str:
        """Set path."""
        return self.set(path_to_str(path))


def path_to_str(path: Path | str) -> str:
    """Convert path to string."""
    return f'"{path}"'


if __name__ == "__main__":
    from rich.markdown import Markdown as Md

    from slurmbench import root_logging

    v = Variable("foo")
    p = Path("tmp/there is a space/file.txt")
    sh_lines = "\n".join(
        [v.set(path_to_str(p)), f"echo {v.eval()}"],
    )
    root_logging.CONSOLE.print(Md(f"```bash\n{sh_lines}\n```"))



"""Prelude topic visitor module."""

# Due to typer usage:
# ruff: noqa: PLC0414

from slurmbench.tool.description import Description as Description
from slurmbench.topic.visitor import Tools as Tools



"""Experience monitors."""

from __future__ import annotations

from enum import StrEnum
from typing import TYPE_CHECKING, final

import slurmbench.samples.items as smp
import slurmbench.samples.status as smp_status
from slurmbench import tab_files

if TYPE_CHECKING:
    from collections.abc import Iterable, Iterator

    from . import file_system as exp_fs


class UnresolvedSamplesTSVHeader(StrEnum):
    """Unresolved samples TSV header."""

    ROW_NUMBER = "row_number"
    SAMPLE_UID = "sample_uid"


@final
class UnresolvedSamplesTSVReader(
    tab_files.TSVReader[UnresolvedSamplesTSVHeader, smp.RowNumbered],
):
    """Unresolved samples TSV reader."""

    @classmethod
    def header_type(cls) -> type[UnresolvedSamplesTSVHeader]:
        """Get header."""
        return UnresolvedSamplesTSVHeader

    def __iter__(self) -> Iterator[smp.RowNumbered]:
        """Iterate over unresolved samples."""
        for row in self._csv_reader:
            row_number = int(self._get_cell(row, self.header_type().ROW_NUMBER))
            sample_uid = self._get_cell(row, self.header_type().SAMPLE_UID)
            yield smp.RowNumbered(row_number, sample_uid)


@final
class UnresolvedSamplesTSVWriter(
    tab_files.TSVWriter[UnresolvedSamplesTSVHeader, smp.RowNumbered],
):
    """Unresolved samples TSV writer."""

    @classmethod
    def header_type(cls) -> type[UnresolvedSamplesTSVHeader]:
        """Get header."""
        return UnresolvedSamplesTSVHeader

    @classmethod
    def reader_type(cls) -> type[UnresolvedSamplesTSVReader]:
        """Get reader."""
        return UnresolvedSamplesTSVReader

    def _to_cell(
        self,
        item: smp.RowNumbered,
        column_id: UnresolvedSamplesTSVHeader,
    ) -> object:
        """Get cell from item."""
        match column_id:
            case UnresolvedSamplesTSVHeader.ROW_NUMBER:
                return item.row_number()
            case UnresolvedSamplesTSVHeader.SAMPLE_UID:
                return item.uid()


class ResolvedSample:
    """Resolved sample."""

    def __init__(
        self,
        sample_uid: str,
        status: smp_status.Status,
    ) -> None:
        """Initialize."""
        self.__sample_uid = sample_uid
        self.__status = status

    def sample_uid(self) -> str:
        """Get sample UID."""
        return self.__sample_uid

    def status(self) -> smp_status.Status:
        """Get satus."""
        return self.__status


class ResolvedSamplesTSVHeader(StrEnum):
    """Resolved samples TSV header."""

    SAMPLE_UID = "sample_uid"
    STATUS = "status"


@final
class ResolvedSampleTSVReader(
    tab_files.TSVReader[ResolvedSamplesTSVHeader, ResolvedSample],
):
    """Error samples TSV reader."""

    @classmethod
    def header_type(cls) -> type[ResolvedSamplesTSVHeader]:
        """Get header type."""
        return ResolvedSamplesTSVHeader

    def __iter__(self) -> Iterator[ResolvedSample]:
        """Iterate over resolved samples."""
        for row in self._csv_reader:
            sample_uid = self._get_cell(row, self.header_type().SAMPLE_UID)
            status = smp_status.from_str(
                self._get_cell(row, self.header_type().STATUS),
            )
            yield ResolvedSample(sample_uid, status)


@final
class ResolvedSamplesTSVWriter(
    tab_files.TSVWriter[ResolvedSamplesTSVHeader, ResolvedSample],
):
    """Resolved samples TSV writer."""

    @classmethod
    def header_type(cls) -> type[ResolvedSamplesTSVHeader]:
        """Get header type."""
        return ResolvedSamplesTSVHeader

    @classmethod
    def reader_type(cls) -> type[ResolvedSampleTSVReader]:
        """Get reader."""
        return ResolvedSampleTSVReader

    def _to_cell(
        self,
        item: ResolvedSample,
        column_id: ResolvedSamplesTSVHeader,
    ) -> object:
        """Get cell from item."""
        match column_id:
            case ResolvedSamplesTSVHeader.SAMPLE_UID:
                return item.sample_uid()
            case ResolvedSamplesTSVHeader.STATUS:
                return item.status()


def write_unresolved_samples(
    work_fs_manager: exp_fs.WorkManager,
    samples: Iterable[smp.RowNumbered],
) -> None:
    """Write unresolved samples."""
    with UnresolvedSamplesTSVWriter.auto_open(
        work_fs_manager.unresolved_samples_tsv(),
    ) as writer:
        writer.write_bunch(samples)


def write_resolved_samples(
    work_fs_manager: exp_fs.WorkManager,
    samples_with_status: Iterable[tuple[smp.RowNumbered, smp_status.Status]],
) -> None:
    """Write resolved samples."""
    with ResolvedSamplesTSVWriter.auto_open(
        work_fs_manager.resolved_samples_tsv(),
    ) as writer:
        writer.write_bunch(
            ResolvedSample(sample.uid(), status)
            for sample, status in samples_with_status
        )


def update_samples_resolution_status(
    work_fs_manager: exp_fs.WorkManager,
    samples_with_status: Iterable[tuple[smp.RowNumbered, smp_status.Status]],
) -> None:
    """Move resolved samples from unresolved file to resolved file."""
    samples_with_status = list(samples_with_status)
    write_resolved_samples(work_fs_manager, samples_with_status)

    with UnresolvedSamplesTSVReader.open(
        work_fs_manager.unresolved_samples_tsv(),
    ) as reader:
        resolved_sample_row_numbers = {
            sample.row_number() for sample, _ in samples_with_status
        }
        new_unresolved_samples = [
            sample
            for sample in reader
            if sample.row_number() not in resolved_sample_row_numbers
        ]

    with UnresolvedSamplesTSVWriter.open(
        work_fs_manager.unresolved_samples_tsv(),
        "w",
    ) as writer:
        writer.write_bunch(new_unresolved_samples)


def iter_unresolved_samples(
    work_fs_manager: exp_fs.WorkManager,
) -> Iterator[smp.RowNumbered]:
    """Iterate over unresolved samples."""
    with UnresolvedSamplesTSVReader.open(
        work_fs_manager.unresolved_samples_tsv(),
    ) as reader:
        yield from reader


def iter_resolved_samples(
    work_fs_manager: exp_fs.WorkManager,
) -> Iterator[ResolvedSample]:
    """Iterate over resolved samples."""
    with ResolvedSampleTSVReader.open(
        work_fs_manager.resolved_samples_tsv(),
    ) as reader:
        yield from reader



"""Tool description abstract module."""

import slurmbench.topic.description as topic_desc


class Description:
    """Tool description."""

    def __init__(
        self,
        name: str,
        cmd: str,
        topic_description: topic_desc.Description,
    ) -> None:
        self.__name = name
        self.__cmd = cmd
        self.__topic_description = topic_description

    def name(self) -> str:
        """Get name."""
        return self.__name

    def cmd(self) -> str:
        """Get command."""
        return self.__cmd

    def topic(self) -> topic_desc.Description:
        """Get topic description."""
        return self.__topic_description



"""Experiment input/output logics."""

from __future__ import annotations

from pathlib import Path
from typing import final

import slurmbench.samples.file_system as smp_fs
import slurmbench.samples.items as smp
import slurmbench.tool.description as tool_desc

from .bash import file_system as exp_bash_fs
from .slurm import file_system as exp_slurm_fs


class ManagerBase:
    """Experiment file system manager base."""

    CONFIG_YAML_NAME = Path("config.yaml")
    IN_PROGRESS_YAML_NAME = Path("in_progress.yaml")

    SCRIPT_DIR_NAME = Path("scripts")

    def __init__(
        self,
        root_directory_path: Path,
        tool_description: tool_desc.Description,
        experiment_name: str,
    ) -> None:
        """Initialize."""
        self._root_directory_path = root_directory_path
        self._tool_description = tool_description
        self._experiment_name = experiment_name

    def tool_description(self) -> tool_desc.Description:
        """Get tool description."""
        return self._tool_description

    def experiment_name(self) -> str:
        """Get experiment name."""
        return self._experiment_name

    def root_dir(self) -> Path:
        """Get root directory path."""
        return self._root_directory_path

    def topic_dir(self) -> Path:
        """Get topic directory path."""
        return self._root_directory_path / self._tool_description.topic().name()

    def tool_dir(self) -> Path:
        """Get tool directory path."""
        return self.topic_dir() / self._tool_description.name()

    def exp_dir(self) -> Path:
        """Get experiment directory path."""
        return self.tool_dir() / self._experiment_name

    #
    # Experiment files
    #
    def config_yaml(self) -> Path:
        """Get config file."""
        return self.exp_dir() / self.CONFIG_YAML_NAME

    def in_progress_yaml(self) -> Path:
        """Get in progress file."""
        return self.exp_dir() / self.IN_PROGRESS_YAML_NAME

    #
    # Sbatch scripts
    #
    def scripts_fs_manager(self, date_str: str) -> exp_bash_fs.Manager:
        """Get experiment scripts file system manager."""
        return exp_bash_fs.Manager(self.exp_dir() / self.SCRIPT_DIR_NAME, date_str)

    #
    # Sample experiment directories
    #
    def sample_dir(self, sample_dirname: str | Path) -> Path:
        """Get sample experiment directory path."""
        return self.exp_dir() / sample_dirname

    def sample_fs_manager(self, sample: smp.RowNumbered) -> smp_fs.Manager:
        """Get sample experiment directory path."""
        return smp_fs.Manager.from_exp_dir_and_sample(self.exp_dir(), sample)


@final
class DataManager(ManagerBase):
    """Data experiment manager."""

    SAMPLES_TSV_NAME = Path("samples.tsv")

    TOOL_ENV_WRAPPER_SCRIPT_NAME = Path("env_wrapper.sh")

    ERRORS_TSV_NAME = Path("errors.tsv")
    HISTORY_YAML_NAME = Path("history.yaml")

    def samples_tsv(self) -> Path:
        """Get samples TSV file."""
        return self.root_dir() / self.SAMPLES_TSV_NAME

    def tool_env_script_sh(self) -> Path:
        """Get tool environment script file."""
        return self.tool_dir() / self.TOOL_ENV_WRAPPER_SCRIPT_NAME

    def errors_tsv(self) -> Path:
        """Get errors file."""
        return self.exp_dir() / self.ERRORS_TSV_NAME

    def history_yaml(self) -> Path:
        """Get history file."""
        return self.exp_dir() / self.HISTORY_YAML_NAME


@final
class WorkManager(ManagerBase):
    """Work experiment manager."""

    UNRESOLVED_SAMPLES_TSV_NAME = Path("unresolved_samples.tsv")
    RESOLVED_SAMPLES_TSV_NAME = Path("resolved_samples.tsv")

    TMP_SLURM_LOG_DIR_NAME = Path("logs")

    def __init__(
        self,
        root_directory_path: Path,
        tool_description: tool_desc.Description,
        experiment_name: str,
    ) -> None:
        super().__init__(root_directory_path, tool_description, experiment_name)
        self.__slurm_log_fs_manager = exp_slurm_fs.LogsManager(
            self.exp_dir() / self.TMP_SLURM_LOG_DIR_NAME,
        )

    def unresolved_samples_tsv(self) -> Path:
        """Get unresolved samples TSV file."""
        return self.exp_dir() / self.UNRESOLVED_SAMPLES_TSV_NAME

    def resolved_samples_tsv(self) -> Path:
        """Get resolved samples TSV file."""
        return self.exp_dir() / self.RESOLVED_SAMPLES_TSV_NAME

    #
    # Tmp sbatch logs
    #
    def slurm_log_fs_manager(self) -> exp_slurm_fs.LogsManager:
        """Get slurm logs manager."""
        return self.__slurm_log_fs_manager


def data_and_working_managers(
    data_dir: Path,
    working_dir: Path,
    tool_description: tool_desc.Description,
    experiment_name: str,
) -> tuple[DataManager, WorkManager]:
    """Get data and working managers."""
    return (
        DataManager(data_dir, tool_description, experiment_name),
        WorkManager(working_dir, tool_description, experiment_name),
    )



"""Help application."""

# Due to typer usage:
# ruff: noqa: FBT002

import logging
from pathlib import Path
from typing import Annotated

import typer
from rich.markdown import Markdown as Md

import slurmbench.help.tool_tree as help_tool_tree
import slurmbench.topics.visitor as topics_visitor
from slurmbench import root_logging

_LOGGER = logging.getLogger(__name__)


def new(topics_type: type[topics_visitor.Topics]) -> typer.Typer:
    """Create a new help application."""
    app = typer.Typer(name="help", help="Help commands", rich_markup_mode="rich")
    tool_tree_app = ToolTree(topics_type)
    app.command(name=tool_tree_app.NAME, help=tool_tree_app.help())(
        tool_tree_app.main,
    )
    return app


class ToolTree[T: type[topics_visitor.Topics]]:
    """Help application."""

    class Opts:
        """Tool tree options."""

        MARKDOWN = typer.Argument(help="Markdown file")

    NAME = "tool-tree"

    def __init__(self, topics_type: T) -> None:
        """Initialize."""
        self._topics_type = topics_type

    def topics_type(self) -> T:
        """Get topics type."""
        return self._topics_type

    def main(
        self,
        markdown_file: Annotated[Path | None, Opts.MARKDOWN] = None,
        debug: Annotated[bool, root_logging.OPT_DEBUG] = False,
    ) -> None:
        """Generate the tree of available tools."""
        root_logging.init_logger(_LOGGER, "Running tool tree helper", debug)
        md_str = help_tool_tree.tool_tree(self.topics_type())

        root_logging.CONSOLE.print(Md(md_str))

        if markdown_file is not None:
            _LOGGER.info("Saving markdown file to `%s`", markdown_file)
            with markdown_file.open("w") as f_out:
                f_out.write(md_str)

    def help(self) -> str:
        """Get help."""
        return "Generate the tree of available tools"



"""Tool connector prelude."""

# Due to typer usage:
# ruff: noqa: PLC0414

from slurmbench.tool.connector import Arg as Arg
from slurmbench.tool.connector import Arguments as Arguments
from slurmbench.tool.connector import Names as Names
from slurmbench.tool.connector import OnlyOptions as OnlyOptions
from slurmbench.tool.connector import WithArguments as WithArguments
from slurmbench.tool.description import Description as Description
from slurmbench.topic.visitor import Tools as Tools



"""Experiment managers."""

from __future__ import annotations

from typing import TYPE_CHECKING, final

import slurmbench.tool.connector as tool_connector

if TYPE_CHECKING:
    from . import file_system as exp_fs


class WithOptions[C: tool_connector.WithOptions]:
    """Experiment manager."""

    def __init__(
        self,
        exp_name: str,
        data_exp_fs_manager: exp_fs.DataManager,
        work_exp_fs_manager: exp_fs.WorkManager,
        tool_connector: C,
    ) -> None:
        self._exp_name = exp_name
        self._data_exp_fs_manager = data_exp_fs_manager
        self._work_exp_fs_manager = work_exp_fs_manager
        self._tool_connector = tool_connector

    def exp_name(self) -> str:
        """Get experiment name."""
        return self._exp_name

    def data_fs_manager(self) -> exp_fs.DataManager:
        """Get data file system manager."""
        return self._data_exp_fs_manager

    def work_fs_manager(self) -> exp_fs.WorkManager:
        """Get working file system manager."""
        return self._work_exp_fs_manager

    def tool_connector(self) -> C:
        """Get tool connector."""
        return self._tool_connector


@final
class OnlyOptions(WithOptions[tool_connector.OnlyOptions]):
    """Experiment manager with only options."""


@final
class WithArguments[N: tool_connector.Names](
    WithOptions[tool_connector.WithArguments[N]],
):
    """Experiment manager with arguments."""



"""Experiment history logic."""

from __future__ import annotations

from typing import Any, Self, final

import slurmbench.samples.status as smp_status
from slurmbench.yaml_interface import YAMLInterface

from . import in_progress, monitors, stats
from . import managers as exp_managers


@final
class Event(YAMLInterface):
    """Base class for event logics."""

    KEY_DATE = "date"
    KEY_JOB_ID = "job_id"
    KEY_STATS = "stats"

    @classmethod
    def from_yaml_load(cls, obj_dict: dict[str, Any]) -> Self:
        """Convert dict to object."""
        return cls(
            obj_dict[cls.KEY_DATE],
            obj_dict.get(cls.KEY_JOB_ID),
            stats.Samples.from_yaml_load(obj_dict[cls.KEY_STATS]),
        )

    def __init__(self, date: str, job_id: str | None, stats: stats.Samples) -> None:
        self._date = date
        self._job_id = job_id
        self._stats = stats

    def date(self) -> str:
        """Get date."""
        return self._date

    def job_id(self) -> str | None:
        """Get job id."""
        return self._job_id

    def stats(self) -> stats.Samples:
        """Get stats."""
        return self._stats

    def to_yaml_dump(self) -> dict[str, Any]:
        """Convert to dict."""
        return {
            self.KEY_DATE: self._date,
            self.KEY_JOB_ID: self._job_id,
            self.KEY_STATS: self._stats.to_yaml_dump(),
        }


@final
class History(list[Event], YAMLInterface):
    """Experiment run history."""

    @classmethod
    def from_yaml_load(cls, obj_list: list[dict[str, Any]]) -> Self:
        """Convert list to object."""
        return cls(
            [Event.from_yaml_load(event_dict) for event_dict in obj_list],
        )

    def to_yaml_dump(self) -> list[dict[str, Any]]:
        """Convert to dict."""
        return [event.to_yaml_dump() for event in self]


def update_history(
    exp_manager: exp_managers.OnlyOptions | exp_managers.WithArguments,
) -> Event:
    """Update history."""
    event = Event(*_date_and_job_id(exp_manager), _stats_from_monitors(exp_manager))
    _add_event_to_history_yaml(exp_manager, event)
    return event


def _date_and_job_id(
    exp_manager: exp_managers.OnlyOptions | exp_managers.WithArguments,
) -> tuple[str, str | None]:
    if exp_manager.work_fs_manager().in_progress_yaml().exists():
        in_progress_metadata = in_progress.InWorkingDirectory.from_yaml(
            exp_manager.work_fs_manager().in_progress_yaml(),
        )
        return in_progress_metadata.date(), in_progress_metadata.job_id()
    return in_progress.get_today_format_string(), None


def _stats_from_monitors(
    exp_manager: exp_managers.OnlyOptions | exp_managers.WithArguments,
) -> stats.Samples:
    number_of_status = smp_status.StatusMap.default(0)
    for resolved_sample in monitors.iter_resolved_samples(
        exp_manager.work_fs_manager(),
    ):
        number_of_status[resolved_sample.status()] += 1

    number_of_status[smp_status.Error.NOT_RUN] = sum(
        1 for _ in monitors.iter_unresolved_samples(exp_manager.work_fs_manager())
    )

    return stats.Samples.from_status_map(number_of_status)


def _add_event_to_history_yaml(
    exp_manager: exp_managers.OnlyOptions | exp_managers.WithArguments,
    event: Event,
) -> None:
    if exp_manager.data_fs_manager().history_yaml().exists():
        history = History.from_yaml(exp_manager.data_fs_manager().history_yaml())
    else:
        history = History()
    history.append(event)
    history.to_yaml(exp_manager.data_fs_manager().history_yaml())



"""Bash experiment file system."""

from __future__ import annotations

from enum import StrEnum
from pathlib import Path

import slurmbench.experiment.bash.items as exp_bash_items


class Manager:
    """Bash experiment file system manager."""

    def __init__(self, script_dir: Path, date_str: str) -> None:
        """Inititialize."""
        self.__script_dir = script_dir
        self.__date_str = date_str

    def scripts_dir(self) -> Path:
        """Get script directory path."""
        return self.__script_dir

    def sbatch_script(self) -> Path:
        """Get the sbatch script file path."""
        return self.scripts_dir() / f"{self.__date_str}_sbatch.sh"

    def step_script(self, step: exp_bash_items.Steps) -> Path:
        """Get bash step status file."""
        return self.scripts_dir() / ScriptStepFilesBuilder.script_step(
            self.__date_str,
            step,
        )


class ScriptStepFilesBuilder:
    """Script step files builder."""

    EXT = "sh"

    # REFACTOR use common part with logs
    class CommandStepsPrefix(StrEnum):
        """Command steps."""

        INIT_ENV = "init_env"
        COMMAND = "command"
        CLOSE_ENV = "close_env"

        @classmethod
        def from_step(
            cls,
            step: exp_bash_items.Steps,
        ) -> ScriptStepFilesBuilder.CommandStepsPrefix:
            """Convert bash step to command step prefix."""
            match step:
                case exp_bash_items.Steps.INIT_ENV:
                    return cls.INIT_ENV
                case exp_bash_items.Steps.COMMAND:
                    return cls.COMMAND
                case exp_bash_items.Steps.CLOSE_ENV:
                    return cls.CLOSE_ENV

    @classmethod
    def script_step(cls, date_str: str, step: exp_bash_items.Steps) -> Path:
        """Get command step status filename."""
        return Path(f"{date_str}_{cls.CommandStepsPrefix.from_step(step)}.{cls.EXT}")



"""Topics abstract module."""

from __future__ import annotations


class Description:
    """Topic description."""

    def __init__(self, name: str, cmd: str) -> None:
        self.__name = name
        self.__cmd = cmd

    def name(self) -> str:
        """Get name."""
        return self.__name

    def cmd(self) -> str:
        """Get command."""
        return self.__cmd



"""Tabular files."""

import _csv
import csv
from abc import ABC, abstractmethod
from collections.abc import Generator, Iterable, Iterator
from contextlib import contextmanager
from enum import StrEnum
from pathlib import Path
from typing import Literal, Self, final

type WriteMode = Literal["w", "a"]


class HeaderMap:
    """Header map."""

    def __init__(self, col_id: list[str]) -> None:
        self.__col_id = col_id
        self.__col_idx = {name: idx for idx, name in enumerate(col_id)}

    def names(self) -> list[str]:
        """Get column names."""
        return self.__col_id

    def indices(self) -> dict[str, int]:
        """Get column indices."""
        return self.__col_idx

    def id_from_idx(self, idx: int) -> str:
        """Get column ID from index."""
        return self.__col_id[idx]

    def idx_from_id(self, name: str) -> int:
        """Get column index from ID."""
        return self.__col_idx[name]


class Delimiter(ABC):
    """Tabular file delimiter."""

    @classmethod
    @abstractmethod
    def char(cls) -> str:
        """Get delimiter character."""
        raise NotImplementedError


class _Base[H: StrEnum, D: Delimiter](ABC):
    """Reader and writer base class."""

    @classmethod
    @abstractmethod
    def header_type(cls) -> type[H]:
        """Get header."""
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def delimiter(cls) -> type[D]:
        """Get delimiter type."""
        raise NotImplementedError

    def __init__(self, file: Path, header_map: HeaderMap) -> None:
        self._file = file
        self._header_map = header_map

    def file(self) -> Path:
        """Get file."""
        return self._file

    def header_map(self) -> HeaderMap:
        """Get header map."""
        return self._header_map


class Reader[H: StrEnum, D: Delimiter, T](_Base[H, D]):
    """Tabular file reader."""

    @classmethod
    @contextmanager
    def open(cls, file: Path) -> Generator[Self]:
        """Open TSV file for reading."""
        with file.open() as f_in:
            reader = cls(file, csv.reader(f_in, delimiter=cls.delimiter().char()))
            yield reader

    def __init__(self, file: Path, csv_reader: _csv.Reader) -> None:
        """Initialize object."""
        self._csv_reader = csv_reader
        super().__init__(file, HeaderMap(next(self._csv_reader)))

    @abstractmethod
    def __iter__(self) -> Iterator[T]:
        """Iterate over the items."""
        raise NotImplementedError

    def _get_cell(self, row: list[str], column_id: H) -> str:
        return row[self._header_map.idx_from_id(column_id)]


class Writer[H: StrEnum, D: Delimiter, T](_Base[H, D]):
    """Unresolved samples TSV writer."""

    @classmethod
    @abstractmethod
    def reader_type(cls) -> type[Reader[H, D, T]]:
        """Get reader."""
        raise NotImplementedError

    @classmethod
    @contextmanager
    def auto_open(cls, file: Path) -> Generator[Self]:
        """Open TSV file for writing."""
        mode: WriteMode = "a" if file.exists() else "w"
        with cls.open(file, mode) as writer:
            yield writer

    @classmethod
    @contextmanager
    def open(cls, file: Path, mode: WriteMode) -> Generator[Self]:
        """Open TSV file for writing."""
        match mode:
            case "w":
                header_map = None
            case "a":
                if file.exists():
                    with cls.reader_type().open(file) as reader:
                        header_map = reader.header_map()
                else:
                    header_map = None
        with file.open(mode) as f_out:
            writer = cls(
                file,
                csv.writer(f_out, delimiter=cls.delimiter().char()),
                header_map,
            )
            yield writer

    def __init__(
        self,
        file: Path,
        csv_writer: _csv.Writer,
        header_map: HeaderMap | None,
    ) -> None:
        """Initialize object."""
        self._csv_writer = csv_writer
        super().__init__(
            file,
            (header_map if header_map is not None else self._write_header()),
        )

    def write(self, item: T) -> None:
        """Write row."""
        self._csv_writer.writerow(
            [
                self._to_cell(item, self.header_type()(name))
                for name in self._header_map.names()
            ],
        )

    def write_bunch(self, items: Iterable[T]) -> None:
        """Write bunch of rows."""
        for item in items:
            self.write(item)

    @abstractmethod
    def _to_cell(self, item: T, column_id: H) -> object:
        """Get cell from item."""
        raise NotImplementedError

    def _write_header(self) -> HeaderMap:
        header_row = list(map(str, self.header_type()))
        self._csv_writer.writerow(header_row)
        return HeaderMap(header_row)


@final
class TSVDelimiter(Delimiter):
    """Tabular file delimiter."""

    @classmethod
    def char(cls) -> str:
        """Get delimiter character."""
        return "\t"


class TSVReader[H: StrEnum, T](Reader[H, TSVDelimiter, T]):
    """TSV file reader."""

    @classmethod
    def delimiter(cls) -> type[TSVDelimiter]:
        """Get delimiter."""
        return TSVDelimiter


class TSVWriter[H: StrEnum, T](Writer[H, TSVDelimiter, T]):
    """TSV file writer."""

    @classmethod
    def delimiter(cls) -> type[TSVDelimiter]:
        """Get delimiter."""
        return TSVDelimiter



"""Experiment errors module."""

from __future__ import annotations

from enum import StrEnum
from typing import TYPE_CHECKING, final

import slurmbench.samples.items as smp
import slurmbench.samples.status as smp_status
from slurmbench import tab_files

if TYPE_CHECKING:
    from collections.abc import Iterable, Iterator

    from . import file_system as exp_fs


class SampleError:
    """Sample error."""

    @classmethod
    def sample_with_missing_inputs(
        cls,
        sample_item: smp.RowNumbered,
    ) -> SampleError:
        """Instantiate a sample error with a missing inputs reason."""
        return cls(sample_item.uid(), smp_status.Error.MISSING_INPUTS)

    @classmethod
    def sample_with_error(
        cls,
        sample_item: smp.RowNumbered,
    ) -> SampleError:
        """Instantiate a sample error with an error reason."""
        return cls(sample_item.uid(), smp_status.Error.ERROR)

    def __init__(self, sample_uid: str, reason: smp_status.Error) -> None:
        """Initialize."""
        self.__sample_uid = sample_uid
        self.__reason = reason

    def sample_uid(self) -> str:
        """Get sample UID."""
        return self.__sample_uid

    def reason(self) -> smp_status.Error:
        """Get reason."""
        return self.__reason


class ErrorsTSVHeader(StrEnum):
    """Error samples TSV header."""

    SAMPLE_UID = "sample_uid"
    REASON = "reason"


@final
class ErrorsTSVReader(tab_files.TSVReader[ErrorsTSVHeader, SampleError]):
    """Error samples TSV reader."""

    @classmethod
    def header_type(cls) -> type[ErrorsTSVHeader]:
        """Get header type."""
        return ErrorsTSVHeader

    def __iter__(self) -> Iterator[SampleError]:
        """Iterate over error samples."""
        for row in self._csv_reader:
            sample_uid = self._get_cell(row, self.header_type().SAMPLE_UID)
            error_status = smp_status.Error(
                self._get_cell(row, self.header_type().REASON),
            )
            yield SampleError(sample_uid, error_status)


@final
class ErrorsTSVWriter(tab_files.TSVWriter[ErrorsTSVHeader, SampleError]):
    """Error samples TSV writer."""

    @classmethod
    def header_type(cls) -> type[ErrorsTSVHeader]:
        """Get header type."""
        return ErrorsTSVHeader

    @classmethod
    def reader_type(cls) -> type[ErrorsTSVReader]:
        """Get reader type."""
        return ErrorsTSVReader

    def _to_cell(self, item: SampleError, column_id: ErrorsTSVHeader) -> object:
        """Get cell from item."""
        match column_id:
            case ErrorsTSVHeader.SAMPLE_UID:
                return item.sample_uid()
            case ErrorsTSVHeader.REASON:
                return item.reason()

    def write_error_sample(self, error_sample: SampleError) -> None:
        """Write error sample."""
        self.write(error_sample)

    def write_error_samples(self, error_samples: Iterable[SampleError]) -> None:
        """Write error samples."""
        for error_sample in error_samples:
            self.write_error_sample(error_sample)


def write_missing_inputs(
    data_fs_manager: exp_fs.DataManager,
    samples_with_missing_inputs: Iterable[smp.RowNumbered],
) -> None:
    """Write experiment missing inputs."""
    with ErrorsTSVWriter.auto_open(data_fs_manager.errors_tsv()) as out_exp_errors:
        out_exp_errors.write_error_samples(
            (
                SampleError.sample_with_missing_inputs(row_numbered_item)
                for row_numbered_item in samples_with_missing_inputs
            ),
        )


def write_errors(
    data_fs_manager: exp_fs.DataManager,
    samples_with_errors: Iterable[smp.RowNumbered],
) -> None:
    """Write experiment errors."""
    with ErrorsTSVWriter.auto_open(data_fs_manager.errors_tsv()) as out_exp_errors:
        out_exp_errors.write_error_samples(
            (
                SampleError.sample_with_error(row_numbered_item)
                for row_numbered_item in samples_with_errors
            ),
        )



"""Experiment complete job logics."""

from __future__ import annotations

import logging
import shutil
import time
from itertools import chain
from typing import TYPE_CHECKING, Self

import rich.progress as rich_prog

import slurmbench.samples.items as smp
import slurmbench.samples.slurm.status as smp_slurm_status
import slurmbench.samples.status as smp_status
import slurmbench.slurm.bash as slurm_bash
from slurmbench import root_logging

from . import errors as exp_errors
from . import history, monitors
from . import managers as exp_managers
from .bash import items as exp_bash_items
from .slurm import checks as exp_slurm_checks
from .slurm import status as exp_slurm_status

if TYPE_CHECKING:
    from collections.abc import Iterable, Iterator

    from slurmbench.slurm import sacct

_LOGGER = logging.getLogger(__name__)


def resolve_running_samples(
    exp_manager: exp_managers.OnlyOptions | exp_managers.WithArguments,
    unresolved_samples: list[smp.RowNumbered],
    array_job_id: str,
) -> None:
    """Complete experiment."""
    _finished_job_deamon(exp_manager, unresolved_samples, array_job_id)
    conclude_experiment(exp_manager)


def _finished_job_deamon(
    exp_manager: exp_managers.OnlyOptions | exp_managers.WithArguments,
    unresolved_samples: list[smp.RowNumbered],
    array_job_id: str,
) -> None:
    """Run finished job deamon."""
    in_running_job_ids = _running_job_ids(array_job_id, unresolved_samples)

    with rich_prog.Progress(console=root_logging.CONSOLE) as progress:
        slurm_running_task = progress.add_task(
            "Slurm running",
            total=len(in_running_job_ids),
        )

        while in_running_job_ids:
            time.sleep(10)

            _tmp_in_running_job_ids, resolved_samples = _get_resolved_samples(
                exp_manager,
                in_running_job_ids,
            )

            _manage_finished_job(resolved_samples, exp_manager)

            progress.update(
                slurm_running_task,
                advance=(len(in_running_job_ids) - len(_tmp_in_running_job_ids)),
            )
            in_running_job_ids = _tmp_in_running_job_ids


def _get_resolved_samples(
    exp_manager: exp_managers.OnlyOptions | exp_managers.WithArguments,
    in_running_job_ids: list[RunningJobID],
) -> tuple[list[RunningJobID], ResolvedSamples]:
    """Get resolved samples."""
    _tmp_in_running_job_ids: list[RunningJobID] = []
    resolved_samples = ResolvedSamples.new()

    for running_job_id in in_running_job_ids:
        sample_status, sacct_state = _get_job_status(
            exp_manager,
            running_job_id,
        )
        match sample_status:
            case smp_status.Success.OK:
                resolved_samples.ok_samples().append(
                    ResolvedJobID.from_running_job_id(
                        running_job_id,
                        sample_status,
                        sacct_state,
                    ),
                )
            case smp_status.Error.ERROR:
                resolved_samples.error_samples().append(
                    ResolvedJobID.from_running_job_id(
                        running_job_id,
                        sample_status,
                        sacct_state,
                    ),
                )
            case smp_status.Error.NOT_RUN:
                _tmp_in_running_job_ids.append(running_job_id)

    return _tmp_in_running_job_ids, resolved_samples


class RunningJobID:
    """Running job ID."""

    @classmethod
    def from_array_job_id(
        cls,
        array_job_id: str,
        row_numbered_item: smp.RowNumbered,
    ) -> Self:
        """Get running job ID from array job ID and row numbered item."""
        return cls(
            slurm_bash.array_task_job_id(
                array_job_id,
                str(row_numbered_item.to_base_one()),
            ),
            row_numbered_item,
        )

    def __init__(self, job_id: str, row_numbered_item: smp.RowNumbered) -> None:
        self.__job_id = job_id
        self.__row_numbered_item = row_numbered_item

    def job_id(self) -> str:
        """Get job id."""
        return self.__job_id

    def row_numbered_item(self) -> smp.RowNumbered:
        """Get row numbered item."""
        return self.__row_numbered_item


def _running_job_ids(
    array_job_id: str,
    unresolved_samples: list[smp.RowNumbered],
) -> list[RunningJobID]:
    """Get the list of running job IDs."""
    return [
        RunningJobID.from_array_job_id(array_job_id, running_sample)
        for running_sample in unresolved_samples
    ]


class ResolvedJobID:
    """Resolved job ID."""

    @classmethod
    def from_running_job_id(
        cls,
        running_job_id: RunningJobID,
        status: smp_status.Status,
        sacct_state: sacct.State | None,
    ) -> Self:
        """Get resolved job ID from running job ID and sacct state."""
        return cls(
            running_job_id.job_id(),
            running_job_id.row_numbered_item(),
            status,
            sacct_state,
        )

    def __init__(
        self,
        job_id: str,
        row_numbered_item: smp.RowNumbered,
        status: smp_status.Status,
        sacct_state: sacct.State | None,
    ) -> None:
        self.__job_id = job_id
        self.__row_numbered_item = row_numbered_item
        self.__status = status
        self.__sacct_state = sacct_state

    def job_id(self) -> str:
        """Get job id."""
        return self.__job_id

    def row_numbered_item(self) -> smp.RowNumbered:
        """Get row numbered item."""
        return self.__row_numbered_item

    def status(self) -> smp_status.Status:
        """Get status."""
        return self.__status

    def sacct_state(self) -> sacct.State | None:
        """Get sacct state."""
        return self.__sacct_state


def _get_job_status(
    exp_manager: exp_managers.OnlyOptions | exp_managers.WithArguments,
    running_job_id: RunningJobID,
) -> tuple[smp_status.Status, sacct.State | None]:
    """Get sample experiment status from job id."""
    sacct_states = slurm_bash.get_states([running_job_id.job_id()])
    #
    # Unknown sacct state
    #
    if running_job_id.job_id() not in sacct_states:
        # The job terminated with a success
        if (
            exp_manager.work_fs_manager()
            .slurm_log_fs_manager()
            .script_step_status_file(
                running_job_id.job_id(),
                exp_bash_items.Steps.CLOSE_ENV,
                exp_slurm_status.ScriptSteps.OK,
            )
            .exists()
        ):
            return smp_status.Success.OK, None
        # The job did not terminate or with an error
        return smp_status.Error.ERROR, None
    return smp_status.from_sacct_state(
        sacct_states[running_job_id.job_id()],
    ), sacct_states[running_job_id.job_id()]


class ResolvedSamples:
    """Resolved samples."""

    @classmethod
    def new(cls) -> Self:
        """Get new resolved samples."""
        return cls([], [])

    def __init__(
        self,
        ok_samples: Iterable[ResolvedJobID],
        error_samples: Iterable[ResolvedJobID],
    ) -> None:
        self.__ok_samples = list(ok_samples)
        self.__error_samples = list(error_samples)

    def ok_samples(self) -> list[ResolvedJobID]:
        """Get ok samples."""
        return self.__ok_samples

    def error_samples(self) -> list[ResolvedJobID]:
        """Get error samples."""
        return self.__error_samples

    def __iter__(self) -> Iterator[ResolvedJobID]:
        """Get iterator."""
        return chain(self.__ok_samples, self.__error_samples)


def _manage_finished_job(
    resolved_samples: ResolvedSamples,
    exp_manager: exp_managers.OnlyOptions | exp_managers.WithArguments,
) -> None:
    _manage_finished_ok_jobs(resolved_samples, exp_manager)
    _manage_finished_error_jobs(resolved_samples, exp_manager)

    for resolved_job_id in resolved_samples:
        _move_slurm_logs_to_work_sample_dir(resolved_job_id, exp_manager)
        _move_work_sample_dir_to_data_dir(resolved_job_id, exp_manager)

    monitors.update_samples_resolution_status(
        exp_manager.work_fs_manager(),
        (
            (resolved_job_id.row_numbered_item(), resolved_job_id.status())
            for resolved_job_id in resolved_samples
        ),
    )


def _manage_finished_ok_jobs(
    resolved_samples: ResolvedSamples,
    exp_manager: exp_managers.OnlyOptions | exp_managers.WithArguments,
) -> None:
    for resolved_job_id in resolved_samples.ok_samples():
        slurm_stdout = (
            exp_manager.work_fs_manager()
            .slurm_log_fs_manager()
            .stdout(resolved_job_id.job_id())
        )
        sample_fs_manager = exp_manager.work_fs_manager().sample_fs_manager(
            resolved_job_id.row_numbered_item(),
        )
        sample_fs_manager.sample_dir().mkdir(parents=True, exist_ok=True)
        shutil.copy(slurm_stdout, sample_fs_manager.done_log())


def _manage_finished_error_jobs(
    resolved_samples: ResolvedSamples,
    exp_manager: exp_managers.OnlyOptions | exp_managers.WithArguments,
) -> None:
    for error_job_id in resolved_samples.error_samples():
        slurm_stderr = (
            exp_manager.work_fs_manager()
            .slurm_log_fs_manager()
            .stderr(error_job_id.job_id())
        )
        sample_fs_manager = exp_manager.work_fs_manager().sample_fs_manager(
            error_job_id.row_numbered_item(),
        )
        sample_fs_manager.sample_dir().mkdir(parents=True, exist_ok=True)
        shutil.copy(slurm_stderr, sample_fs_manager.errors_log())

    if resolved_samples.error_samples():
        exp_errors.write_errors(
            exp_manager.data_fs_manager(),
            (
                error_job_id.row_numbered_item()
                for error_job_id in resolved_samples.error_samples()
            ),
        )


def _move_slurm_logs_to_work_sample_dir(
    resolved_job_id: ResolvedJobID,
    exp_manager: exp_managers.OnlyOptions | exp_managers.WithArguments,
) -> None:
    sample_fs_manager = exp_manager.work_fs_manager().sample_fs_manager(
        resolved_job_id.row_numbered_item(),
    )
    smp_slurm_fs_manager = sample_fs_manager.slurm_fs_manager()

    smp_slurm_fs_manager.slurm_dir().mkdir(parents=True, exist_ok=True)

    sacct_state = resolved_job_id.sacct_state()
    if sacct_state is not None:
        smp_slurm_fs_manager.job_state_file_builder().path(sacct_state).touch()

    slurm_bash.write_slurm_stats(
        resolved_job_id.job_id(),
        smp_slurm_fs_manager.stats_psv(),
    )

    slurm_stdout = (
        exp_manager.work_fs_manager()
        .slurm_log_fs_manager()
        .stdout(resolved_job_id.job_id())
    )
    shutil.copy(slurm_stdout, smp_slurm_fs_manager.stdout_log())

    slurm_stderr = (
        exp_manager.work_fs_manager()
        .slurm_log_fs_manager()
        .stderr(resolved_job_id.job_id())
    )
    shutil.copy(slurm_stderr, smp_slurm_fs_manager.stderr_log())

    _command_steps_process_from_slurm_logs(exp_manager, resolved_job_id).to_yaml(
        smp_slurm_fs_manager.command_steps_status_file_manager().path(),
    )


def _command_steps_process_from_slurm_logs(
    exp_manager: exp_managers.OnlyOptions | exp_managers.WithArguments,
    resolved_job_id: ResolvedJobID,
) -> smp_slurm_status.CommandStepsProcess:
    return smp_slurm_status.CommandStepsProcess(
        exp_slurm_checks.script_step_status(
            exp_manager.work_fs_manager(),
            resolved_job_id.job_id(),
            exp_bash_items.Steps.INIT_ENV,
        ),
        exp_slurm_checks.script_step_status(
            exp_manager.work_fs_manager(),
            resolved_job_id.job_id(),
            exp_bash_items.Steps.COMMAND,
        ),
        exp_slurm_checks.script_step_status(
            exp_manager.work_fs_manager(),
            resolved_job_id.job_id(),
            exp_bash_items.Steps.CLOSE_ENV,
        ),
    )


def _move_work_sample_dir_to_data_dir(
    resolved_job_id: ResolvedJobID,
    exp_manager: exp_managers.OnlyOptions | exp_managers.WithArguments,
) -> None:
    work_sample_fs_manager = exp_manager.work_fs_manager().sample_fs_manager(
        resolved_job_id.row_numbered_item(),
    )
    data_sample_fs_manager = exp_manager.data_fs_manager().sample_fs_manager(
        resolved_job_id.row_numbered_item(),
    )
    shutil.rmtree(data_sample_fs_manager.sample_dir(), ignore_errors=True)
    shutil.copytree(
        work_sample_fs_manager.sample_dir(),
        data_sample_fs_manager.sample_dir(),
    )
    shutil.rmtree(work_sample_fs_manager.sample_dir(), ignore_errors=True)


def conclude_experiment(
    exp_manager: exp_managers.OnlyOptions | exp_managers.WithArguments,
) -> None:
    """Conclude the experiment."""
    event = history.update_history(exp_manager)
    _LOGGER.info("Experiment `%s` finished", exp_manager.exp_name())
    _LOGGER.info(
        (
            "Experiment stats:\n"
            "* Total number of samples: %d\n"
            "* Number of successful samples: %d\n"
            "* Number of samples with missing inputs: %d\n"
            "* Number of failed samples: %d\n"
            "* Number of not run samples: %d\n"
        ),
        event.stats().total_number_of_samples(),
        event.stats().number_of_successful_samples(),
        event.stats().number_of_samples_with_missing_inputs(),
        event.stats().number_of_failed_samples(),
        event.stats().number_of_not_run_samples(),
    )
    exp_manager.data_fs_manager().in_progress_yaml().unlink()
    _clean_work_directory(exp_manager)


def _clean_work_directory(
    exp_manager: exp_managers.OnlyOptions | exp_managers.WithArguments,
) -> None:
    """Move work to data."""
    _LOGGER.info("Cleaning work directory")
    shutil.rmtree(exp_manager.work_fs_manager().exp_dir(), ignore_errors=True)
    #
    # Try to remove empty parent directories
    #
    tree_to_remove = [
        exp_manager.work_fs_manager().root_dir(),
        exp_manager.work_fs_manager().topic_dir(),
        exp_manager.work_fs_manager().tool_dir(),
    ]
    last_empty = True
    while tree_to_remove and last_empty:
        dir_to_remove = tree_to_remove.pop()
        if not any(dir_to_remove.iterdir()):
            dir_to_remove.rmdir()
        else:
            last_empty = False



"""Experiment bash manager."""

from __future__ import annotations

from typing import TYPE_CHECKING

import slurmbench.experiment.in_progress as exp_in_progress
import slurmbench.experiment.managers as exp_managers

if TYPE_CHECKING:
    from . import file_system as fs


class Manager:
    """Experiment bash manager."""

    @classmethod
    def from_exp_fs_manager(
        cls,
        exp_fs_manager: exp_managers.OnlyOptions | exp_managers.WithArguments,
    ) -> Manager:
        """Initialize from experiment file system manager."""
        date_str = exp_in_progress.get_today_format_string()
        return cls(
            date_str,
            exp_fs_manager.data_fs_manager().scripts_fs_manager(date_str),
            exp_fs_manager.work_fs_manager().scripts_fs_manager(date_str),
        )

    def __init__(
        self,
        date_str: str,
        data_sh_fs_manager: fs.Manager,
        work_sh_fs_manager: fs.Manager,
    ) -> None:
        self.__date_str = date_str
        self.__data_sh_fs_manager = data_sh_fs_manager
        self.__work_sh_fs_manager = work_sh_fs_manager

    def date_str(self) -> str:
        """Get date string."""
        return self.__date_str

    def data_sh_fs_manager(self) -> fs.Manager:
        """Get data shell file system manager."""
        return self.__data_sh_fs_manager

    def work_sh_fs_manager(self) -> fs.Manager:
        """Get working shell file system manager."""
        return self.__work_sh_fs_manager



"""Experiment checking module."""

from __future__ import annotations

import logging
from enum import StrEnum
from typing import TYPE_CHECKING, cast

import slurmbench.tool.config as tool_cfg
import slurmbench.tool.connector as tool_connector

from . import file_system as exp_fs
from . import managers as exp_managers

if TYPE_CHECKING:
    from pathlib import Path


_LOGGER = logging.getLogger(__name__)


class RunOK[M: exp_managers.WithOptions]:
    """OK status."""

    def __init__(self, exp_manager: M) -> None:
        self._exp_manager = exp_manager

    def exp_manager(self) -> M:
        """Get experiment manager."""
        return self._exp_manager


class RunErrors(StrEnum):
    """Experiment checks error status before run."""

    NO_PERMISSION = "no_permission"
    READ_CONFIG_FAILED = "read_config_failed"
    MISSING_TOOL_ENV_WRAPPER_SCRIPT = "missing_tool_env_wrapper_script"


def check_before_start(
    exp_name: str,
    data_dir: Path,
    work_dir: Path,
    tool_config_yaml: Path,
    tool_connector_type: type[
        tool_connector.OnlyOptions | tool_connector.WithArguments
    ],
) -> RunOK[exp_managers.OnlyOptions | exp_managers.WithArguments] | RunErrors:
    """Check experiment."""
    match _check_read_write_access(data_dir, work_dir):
        case PermissionErrors():
            return RunErrors.NO_PERMISSION

    connector_or_error = instantiate_connector(tool_connector_type, tool_config_yaml)

    match connector_or_error:
        case RunErrors():
            return connector_or_error

    _LOGGER.debug(
        "Experiment config:\n%s",
        connector_or_error.to_config().to_yaml_dump(),
    )

    exp_manager = _get_exp_manager(exp_name, data_dir, work_dir, connector_or_error)

    if _missing_env_wrapper_script(exp_manager.data_fs_manager()):
        return RunErrors.MISSING_TOOL_ENV_WRAPPER_SCRIPT

    return RunOK(exp_manager)


class PermissionOK(StrEnum):
    """Permission OK status."""

    READ_WRITE = "read_write"


class PermissionErrors(StrEnum):
    """Permission status."""

    NO_READ_ACCESS = "no_read_access"
    NO_WRITE_ACCESS = "no_write_access"


type PermissionStatus = PermissionOK | PermissionErrors


def _check_read_write_access(data_dir: Path, work_dir: Path) -> PermissionStatus:
    """Check read and write access."""
    match status := _check_read_write_access_data(data_dir):
        case PermissionErrors():
            return status

    match status := _check_read_write_access_work(work_dir):
        case PermissionErrors():
            return status

    return PermissionOK.READ_WRITE


def _check_read_write_access_data(data_dir: Path) -> PermissionStatus:
    """Check read and write access."""
    if not data_dir.exists():
        _LOGGER.critical("Data directory %s does not exist", data_dir)
        return PermissionErrors.NO_READ_ACCESS

    file_test = data_dir / "test_read_write.txt"
    try:
        file_test.write_text("test")
    except OSError as err:
        _LOGGER.critical("No write access to %s with exception: %s", data_dir, err)
        return PermissionErrors.NO_WRITE_ACCESS

    try:
        file_test.read_text()
    except OSError as err:
        _LOGGER.critical("No read access to %s with exception: %s", data_dir, err)
        file_test.unlink()
        return PermissionErrors.NO_READ_ACCESS

    file_test.unlink()

    return PermissionOK.READ_WRITE


def _check_read_write_access_work(work_dir: Path) -> PermissionStatus:
    """Check read and write access."""
    try:
        work_dir.mkdir(parents=True, exist_ok=True)
    except OSError:
        _LOGGER.exception("No write access to %s", work_dir)
        return PermissionErrors.NO_WRITE_ACCESS
    file_test = work_dir / "test_read_write.txt"

    try:
        file_test.write_text("test")
    except OSError:
        _LOGGER.exception("No write access to %s", work_dir)
        file_test.unlink(missing_ok=True)
        if not any(work_dir.iterdir()):
            work_dir.rmdir()
        return PermissionErrors.NO_WRITE_ACCESS

    try:
        file_test.read_text()
    except OSError:
        _LOGGER.exception("No read access to %s", work_dir)
        file_test.unlink()
        if not any(work_dir.iterdir()):
            work_dir.rmdir()
        return PermissionErrors.NO_READ_ACCESS

    file_test.unlink()
    if not any(work_dir.iterdir()):
        work_dir.rmdir()

    return PermissionOK.READ_WRITE


# REFACTOR common function so use common ERROR
def instantiate_connector(
    tool_connector_type: type[
        tool_connector.OnlyOptions | tool_connector.WithArguments
    ],
    tool_config_yaml: Path,
) -> tool_connector.OnlyOptions | tool_connector.WithArguments | RunErrors:
    """Instantiate connector."""
    if tool_connector_type is tool_connector.OnlyOptions:
        tool_connector_type = cast(
            "type[tool_connector.OnlyOptions]",
            tool_connector_type,
        )  # Mypy fails to infer otherwise
        return tool_connector_type.from_config(
            tool_cfg.OnlyOptions.from_yaml(tool_config_yaml),
        )
    tool_connector_type = cast(
        "type[tool_connector.WithArguments]",
        tool_connector_type,
    )  # Mypy fails to infer otherwise
    match connector_or_error := tool_connector_type.from_config(
        tool_cfg.WithArguments.from_yaml(tool_config_yaml),
    ):
        case tool_connector.InvalidToolNameError():
            _LOGGER.critical(
                "Invalid tool name `%s` for argument name `%s`."
                " Choose among the valid tools in : {%s}",
                connector_or_error.invalid_tool_name(),
                connector_or_error.arg_name(),
                ", ".join(connector_or_error.valid_tools()),
            )
            return RunErrors.READ_CONFIG_FAILED
        case tool_connector.MissingArgumentNameError():
            _LOGGER.critical(
                "Argument name not found: `%s`."
                " All the argument names must be present: {%s}",
                connector_or_error.missing_arg_name(),
                ", ".join(str(name) for name in connector_or_error.names_type()),
            )
            return RunErrors.READ_CONFIG_FAILED
        case tool_connector.ExtraArgumentNameError():
            _LOGGER.critical(
                "Extra argument name: `%s`."
                " Only argument names in the following set must be present: {%s}",
                connector_or_error.extra_arg_names(),
                ", ".join(str(name) for name in connector_or_error.names_type()),
            )
            return RunErrors.READ_CONFIG_FAILED
    return connector_or_error


def _get_exp_manager(
    exp_name: str,
    data_dir: Path,
    work_dir: Path,
    connector: tool_connector.OnlyOptions | tool_connector.WithArguments,
) -> exp_managers.OnlyOptions | exp_managers.WithArguments:
    data_exp_fs_manager, work_exp_fs_manager = exp_fs.data_and_working_managers(
        data_dir,
        work_dir,
        connector.description(),
        exp_name,
    )
    match connector:
        case tool_connector.OnlyOptions():
            return exp_managers.OnlyOptions(
                exp_name,
                data_exp_fs_manager,
                work_exp_fs_manager,
                connector,
            )
        case tool_connector.WithArguments():
            return exp_managers.WithArguments(
                exp_name,
                data_exp_fs_manager,
                work_exp_fs_manager,
                connector,
            )


def _missing_env_wrapper_script(data_exp_fs_manager: exp_fs.DataManager) -> bool:
    """Check missing env wrapper script."""
    if not data_exp_fs_manager.tool_env_script_sh().exists():
        _LOGGER.critical("Missing tool environment wrapper script")
        _LOGGER.info("You can use `draft-env` command to generate a draft script")
        return True
    return False


class SameExperimentConfigs(StrEnum):
    """Same experiment configs OK status."""

    SAME = "same"


class DifferentExperimentConfigs(StrEnum):
    """Different experiment configs error."""

    DIFFERENT_SYNTAX = "different_syntax"
    NOT_SAME = "not_same"


type ExperimentConfigComparison = SameExperimentConfigs | DifferentExperimentConfigs


def compare_config_vs_config_in_data(
    connector: tool_connector.OnlyOptions | tool_connector.WithArguments,
    config_in_data_yaml: Path,
) -> ExperimentConfigComparison:
    """Compare two experimentation configs."""
    connector_in_data: (
        tool_connector.OnlyOptions
        | tool_connector.WithArguments
        | tool_connector.ArgsLoadError
    )
    match connector:
        case tool_connector.OnlyOptions():
            connector_in_data = tool_connector.OnlyOptions.from_config(
                tool_cfg.OnlyOptions.from_yaml(config_in_data_yaml),
            )
        case tool_connector.WithArguments():
            match connector_in_data := type(connector).from_config(
                tool_cfg.WithArguments.from_yaml(config_in_data_yaml),
            ):
                case (
                    tool_connector.InvalidToolNameError()
                    | tool_connector.MissingArgumentNameError()
                    | tool_connector.ExtraArgumentNameError()
                ):
                    return DifferentExperimentConfigs.DIFFERENT_SYNTAX

    match connector:  # force for type checking
        case tool_connector.OnlyOptions():
            connector_in_data = cast(
                "tool_connector.OnlyOptions",
                connector_in_data,
            )
            is_same = connector.is_same(connector_in_data)
        case tool_connector.WithArguments():
            connector_in_data = cast(
                "tool_connector.WithArguments",
                connector_in_data,
            )
            is_same = connector.is_same(connector_in_data)

    if not is_same:
        _LOGGER.critical(
            "Existing and given experiment configurations are not the same",
        )
        return DifferentExperimentConfigs.NOT_SAME

    return SameExperimentConfigs.SAME



"""Sample bash script logics."""

from collections.abc import Iterator
from enum import StrEnum
from pathlib import Path

import slurmbench.bash.items as bash_items
import slurmbench.experiment.file_system as exp_fs
import slurmbench.samples.file_system as smp_fs
import slurmbench.slurm.bash as slurm_bash


class SampleUIDLinesBuilder:
    """Sample UID bash lines builder."""

    SAMPLES_TSV_VAR = bash_items.Variable("SAMPLES_TSV")

    SAMPLE_UID_VAR = bash_items.Variable("SAMPLE_UID")

    def __init__(self, samples_file: Path) -> None:
        """Initialize."""
        self.__samples_file = samples_file

    def samples_file(self) -> Path:
        """Get samples file."""
        return self.__samples_file

    def lines(self) -> Iterator[str]:
        """Give the bash lines defining the species-sample id variable."""
        yield self.SAMPLES_TSV_VAR.set(bash_items.path_to_str(self.__samples_file))
        yield self.SAMPLE_UID_VAR.set(
            get_sample_attribute(self.__samples_file, smp_fs.TSVHeader.SAMPLE_UID),
        )
        yield f"echo {self.SAMPLE_UID_VAR.eval()}"


def get_sample_attribute(samples_file: Path, attribute: str | StrEnum) -> str:
    """Get sample attribute."""
    # DOCU user can use it to define bash tool commands
    attribute_column_index = smp_fs.columns_name_index(samples_file)[str(attribute)]
    return (
        f"$("
        f'sed -n "{slurm_bash.SLURM_ARRAY_TASK_ID_VAR.eval()}p"'
        f" {SampleUIDLinesBuilder.SAMPLES_TSV_VAR.eval()}"
        f" | cut -f{1 + attribute_column_index}"
        f")"
    )


def sample_shell_fs_manager(exp_fs_manager: exp_fs.ManagerBase) -> smp_fs.Manager:
    """Get sample shell variable file system manager."""
    return smp_fs.Manager(
        exp_fs_manager.sample_dir(SampleUIDLinesBuilder.SAMPLE_UID_VAR.eval()),
    )
